import importlib.util
from typing import Any, Tuple, Union
from packaging import version
import sys
if sys.version_info < (3, 8):
    import importlib_metadata
else:
    import importlib.metadata as importlib_metadata


# TODO: This doesn't work for all packages (`bs4`, `faiss`, etc.) Talk to Sylvain to see how to do with it better.
def is_package_available(pkg_name: str, return_version: bool = False) -> Union[Tuple[bool, str], bool]:
    # Check we're not importing a "pkg_name" directory somewhere but the actual library by trying to grab the version
    package_exists = importlib.util.find_spec(pkg_name) is not None
    package_version = "N/A"
    if package_exists:
        try:
            package_version = importlib.metadata.version(pkg_name)
            package_exists = True
        except importlib.metadata.PackageNotFoundError:
            package_exists = False
        # print(f"Detected {pkg_name} version {package_version}")
    if return_version:
        return package_exists, package_version
    else:
        return package_exists


def is_safetensors_available():
    return is_package_available("safetensors")


def is_sklearn_available():
    return is_package_available("sklearn")


import sys
import collections
import time
import numpy as np
from datetime import datetime
import warnings
from collections import deque
import json
import copy
import os
from torch4keras.snippets import log_info, log_error, log_warn, send_email, set_precision
import math

# 忽略nan的指标
IGNORE_NAN_VALUES = os.environ.get('IGNORE_NAN_VALUES', False)

# 不记录的metrics
SKIP_METRICS = os.environ.get('SKIP_METRICS', {})
SKIP_METRICS = eval(SKIP_METRICS) if isinstance(SKIP_METRICS, str) else SKIP_METRICS

# 不能平滑的指标
NO_SMOOTH_METRICS = os.environ.get('NO_SMOOTH_METRICS', {'lr'})
NO_SMOOTH_METRICS = eval(NO_SMOOTH_METRICS) if isinstance(NO_SMOOTH_METRICS, str) else NO_SMOOTH_METRICS

# 指标的精度
ROUND_PRECISION = int(os.environ.get('ROUND_PRECISION', 4))
_round_precision = eval(f"1e-{ROUND_PRECISION-1}")


def round(v, mode='f'):
    if abs(v) < _round_precision:
        mode = 'e'
    return f"%.{ROUND_PRECISION}{mode}" % v


class CallbackList(object):
    '''把原来在model.py中的callback_fun移植出来，参考Keras的CallbackList重构
    '''
    def __init__(self, callbacks=None, queue_length=10, run_callbacks=True):
        callbacks = callbacks or []
        self.callbacks = [c for c in callbacks]
        self.queue_length = queue_length
        self.run_callbacks = run_callbacks  # 控制全部开启/关闭

    def append(self, callback):
        self.callbacks.append(callback)

    def set_params(self, params):
        for callback in self.callbacks:
            callback.set_params(params)

    def set_trainer(self, trainer):
        for callback in self.callbacks:
            callback.set_trainer(trainer)

    def set_model(self, model):
        for callback in self.callbacks:
            callback.set_model(model)

    def set_optimizer(self, optimizer):
        for callback in self.callbacks:
            callback.set_optimizer(optimizer)

    def set_scheduler(self, scheduler):
        for callback in self.callbacks:
            callback.set_scheduler(scheduler)

    def set_all(self, trainer=None, model=None, optimizer=None, scheduler=None, params=None):
        for callback in self.callbacks:
            callback.set_trainer(trainer)
            callback.set_model(model)
            callback.set_optimizer(optimizer)
            callback.set_scheduler(scheduler)
            callback.set_params(params)

    def on_epoch_begin(self, global_step, epoch, logs=None):
        # 如果是分布式DDP训练，则仅masker_rank可以callback
        if not self.run_callbacks: return
        logs = logs or {}
        for callback in self.callbacks:
            if hasattr(callback, 'run_callback') and (not callback.run_callback): return
            callback.on_epoch_begin(global_step, epoch, logs)
        self._delta_t_batch = 0.
        self._delta_ts_batch_begin = deque([], maxlen=self.queue_length)
        self._delta_ts_batch_end = deque([], maxlen=self.queue_length)

    def on_epoch_end(self, global_step, epoch, logs=None):
        if not self.run_callbacks: return
        logs = logs or {}
        for callback in self.callbacks:
            if hasattr(callback, 'run_callback') and (not callback.run_callback): return
            callback.on_epoch_end(global_step, epoch, logs)

    def on_batch_begin(self, global_step, local_step, logs=None):
        if not self.run_callbacks: return
        logs = logs or {}
        t_before_callbacks = time.time()
        for callback in self.callbacks:
            if hasattr(callback, 'run_callback') and (not callback.run_callback): return
            callback.on_batch_begin(global_step, local_step, logs)
        self._delta_ts_batch_begin.append(time.time() - t_before_callbacks)
        delta_t_median = np.median(self._delta_ts_batch_begin)
        if (self._delta_t_batch > 0. and delta_t_median > 0.95 * self._delta_t_batch and delta_t_median > 0.1):
            warnings.warn(f'Method on_batch_begin() is slow compared to the batch update {delta_t_median}. Check your callbacks.')
        self._t_enter_batch = time.time()

    def on_batch_end(self, global_step, local_step, logs=None):
        if not self.run_callbacks: return
        logs = logs or {}
        if not hasattr(self, '_t_enter_batch'):
            self._t_enter_batch = time.time()
        self._delta_t_batch = time.time() - self._t_enter_batch
        t_before_callbacks = time.time()
        for callback in self.callbacks:
            if hasattr(callback, 'run_callback') and (not callback.run_callback): return
            callback.on_batch_end(global_step, local_step, logs)
        self._delta_ts_batch_end.append(time.time() - t_before_callbacks)
        delta_t_median = np.median(self._delta_ts_batch_end)
        if (self._delta_t_batch > 0. and (delta_t_median > 0.95 * self._delta_t_batch and delta_t_median > 0.1)):
            warnings.warn(f'Method on_batch_end() is slow compared to the batch update {delta_t_median}. Check your callbacks.')

    def on_train_begin(self, logs=None):
        if not self.run_callbacks: return
        logs = logs or {}
        for callback in self.callbacks:
            if hasattr(callback, 'run_callback') and (not callback.run_callback): return
            callback.on_train_begin(logs)

    def on_train_end(self, logs=None):
        if not self.run_callbacks: return
        logs = logs or {}
        for callback in self.callbacks:
            if hasattr(callback, 'run_callback') and (not callback.run_callback): return
            callback.on_train_end(logs)

    def on_dataloader_end(self, logs=None):
        if not self.run_callbacks: return
        logs = logs or {}
        for callback in self.callbacks:
            if hasattr(callback, 'run_callback') and (not callback.run_callback): return
            callback.on_dataloader_end(logs)

    def on_train_step_end(self, logs=None):
        if not self.run_callbacks: return
        logs = logs or {}
        for callback in self.callbacks:
            if hasattr(callback, 'run_callback') and (not callback.run_callback): return
            callback.on_train_step_end(logs)

    def __iter__(self):
        return iter(self.callbacks)


class Callback(object):
    '''Callback基类'''
    def __init__(self, run_callback=True, **kwargs):
        self.trainer = None  # trainer
        self.model = None  # nn.Module模型，或者包含Trainer的nn.Module
        self.optimizer = None  # 优化器
        self.run_callback = run_callback  # 是否运行该callback
    def set_params(self, params):
        self.params = params
    def set_trainer(self, trainer):
        self.trainer = trainer
    def set_model(self, model):
        self.model = model
    def set_optimizer(self, optimizer):
        self.optimizer = optimizer
    def set_scheduler(self, scheduler):
        self.scheduler = scheduler
    def on_train_begin(self, logs=None):
        pass
    def on_train_end(self, logs=None):
        pass
    def on_epoch_begin(self, global_step, epoch, logs=None):
        pass
    def on_epoch_end(self, global_step, epoch, logs=None):
        pass
    def on_batch_begin(self, global_step, local_step, logs=None):
        pass
    def on_batch_end(self, global_step, local_step, logs=None):
        pass
    def on_dataloader_end(self, logs=None):
        pass
    def on_train_step_end(self, logs=None):
        pass


def _process_stateful_metrics(stateful_metrics):
    '''对stateful_metrics进行处理'''
    if stateful_metrics is None:
        stateful_metrics_new = set()
    elif isinstance(stateful_metrics, str):
        stateful_metrics_new = {stateful_metrics}
    elif isinstance(stateful_metrics, (set, tuple, list)):
        stateful_metrics_new = set(stateful_metrics)
    else:
        raise ValueError('Args `stateful_metrics` only support `int/set/tuple/list` format')
    stateful_metrics_new.update(NO_SMOOTH_METRICS)
    return stateful_metrics_new


class SmoothMetric:
    '''指标平滑'''
    def __init__(self, interval=None, stateful_metrics=None) -> None:
        self._values = collections.OrderedDict()  # OrderedDict([('loss', [11.60657262802124, 5]), ('acc', [0.25, 5])])
        self.interval = interval
        self.stateful_metrics = stateful_metrics or set()
        self._seen_so_far = 0
    
    def update(self, current, logs:dict):
        if (self.interval is not None) and (current % self.interval == 1):
            # 如果定义了累积interval，则需要重新累计
            self.reset()
        
        for k, v in logs.items():
            if k in SKIP_METRICS:
                continue
            elif IGNORE_NAN_VALUES and np.isnan(v):
                log_error(f'Value `{k}` at {current} step is nan')
                continue
            elif (k in NO_SMOOTH_METRICS) or (k in self.stateful_metrics):
                self._values[k] = [v, 1]
            elif k not in self._values:
                self._values[k] = [v * (current - self._seen_so_far), current - self._seen_so_far]
            else:
                self._values[k][0] += v * (current - self._seen_so_far)
                self._values[k][1] += (current - self._seen_so_far)
            
        self._seen_so_far = current
        return self._values

    def reset(self):
        self._values = collections.OrderedDict()

    def get_smooth_logs(self, logs=None):
        smooth_logs = {k: v[0]/v[1] for k, v in self._values.items() if k not in SKIP_METRICS}
        
        if logs is not None:
            for k, v in logs.items() :
                if (k in SKIP_METRICS) or (k in smooth_logs):
                    continue
                smooth_logs[k] = v
        return smooth_logs

    def add(self, n, values=None):
        self.update(self._seen_so_far + n, values)


class SmoothMetricsCallback(Callback):
    '''指标平滑的callback，会inplace修改log，影响后续的callback中log
    1）适用情形：希望Logger, Tensorboard，Wandb, EarlyStopping等callbacks中使用的累计平滑的指标
    2）使用方法：初始化后，放在fit()中靠前的位置来对log进行修改
    3）step的平滑是全局来看的（跨epoch不中断），epoch平滑是对每个epoch分别累计计算

    :param interval: int, 平滑时候使用的的step个数
    :param stateful_metrics: list, 以状态量记录指标的格式
    '''
    def __init__(self, interval=100, stateful_metrics=None, verbose=0, **kwargs):
        super(SmoothMetricsCallback, self).__init__(**kwargs)
        self.interval = interval
        self.stateful_metrics = stateful_metrics
        self.smooth_metric_step = SmoothMetric(interval=self.interval, stateful_metrics=self.stateful_metrics)
        if verbose != 0:
            log_info(f'SmoothMetricCallback calculate {interval} steps average metrics')

    def on_epoch_begin(self, global_step, epoch, logs=None):
        self.smooth_metric_epoch = SmoothMetric(stateful_metrics=self.stateful_metrics)

    def on_epoch_end(self, global_step, epoch, logs=None):
        self.smooth_metric_epoch.get_smooth_logs(logs)
        smooth_logs = self.smooth_metric_epoch.get_smooth_logs()
        logs.update(smooth_logs)

    def on_batch_end(self, global_step, local_step, logs=None):
        self.smooth_metric_step.update(global_step+1, logs)
        self.smooth_metric_epoch.update(local_step+1, logs)
        smooth_logs = self.smooth_metric_step.get_smooth_logs()
        logs.update(smooth_logs)


class Progbar(object):
    '''进度条，直接从keras引入'''
    def __init__(self, target, width=30, verbose=1, time_interval=0.05, stateful_metrics=None):
        '''
        :param target: 进度条的step数量
        :param width: 进度条的宽度
        :param verbose: 是否展示进度条
        :param time_interval: 更新进度条的最短时间间隔
        :param stateful_metrics: list, 以状态量记录指标的格式
        '''
        self.target = target
        self.width = width
        self.verbose = verbose
        self.time_interval = time_interval
        self.stateful_metrics = _process_stateful_metrics(stateful_metrics)
        self._dynamic_display = ((hasattr(sys.stdout, 'isatty') and sys.stdout.isatty()) or 'ipykernel' in sys.modules)
        self._total_width = 0
        self._start = time.time()
        self._last_update = 0

    @staticmethod
    def format_time(eta):
        if eta > 3600:
            eta_format = ('%d:%02d:%02d' % (eta // 3600, (eta % 3600) // 60, eta % 60))
        else:
            eta_format = '%d:%02d' % (eta // 60, eta % 60)
        return eta_format

    def update(self, current, values=None):
        '''Updates the progress bar.'''
        now = time.time()
        if self.verbose == 1:
            if (now - self._last_update < self.time_interval and self.target is not None and current < self.target):
                # 训练每个step太快了，则不更新进度条，累计达到一定的时间间隔再更新进度条
                return

            prev_total_width = self._total_width
            if self._dynamic_display:
                sys.stdout.write('\b' * prev_total_width)
                sys.stdout.write('\r')
            else:
                sys.stdout.write('\n')

            if self.target is not None:
                numdigits = int(np.floor(np.log10(self.target))) + 1
                barstr = '%%%dd/%d [' % (numdigits, self.target)
                bar = barstr % current
                prog = float(current) / self.target
                prog_width = int(self.width * prog)
                if prog_width > 0:
                    bar += ('=' * (prog_width - 1))
                    if current < self.target:
                        bar += '>'
                    else:
                        bar += '='
                bar += ('.' * (self.width - prog_width))
                bar += ']'
            else:
                bar = '%7d/Unknown' % current

            self._total_width = len(bar)
            sys.stdout.write(bar)

            if current:
                time_per_unit = (now - self._start) / current
            else:
                time_per_unit = 0
            if self.target is not None and current < self.target:
                eta = time_per_unit * (self.target - current)
                info = ' - ETA: %s' % self.format_time(now - self._start) + '<' + self.format_time(eta)
            else:
                info = ' - %s' % self.format_time(now - self._start)
                if time_per_unit >= 1:
                    info += ' %.0fs/step' % time_per_unit
                elif time_per_unit >= 1e-3:
                    info += ' %.0fms/step' % (time_per_unit * 1e3)
                else:
                    info += ' %.0fus/step' % (time_per_unit * 1e6)

            for k in values:
                info += ' - %s:' % k
                if isinstance(values[k], list):
                    avg = np.mean(values[k][0] / max(1, values[k][1]))  # 指标平滑
                    if abs(avg) > 1e-3:
                        info += f' %.{ROUND_PRECISION}f' % avg
                    else:
                        info += f' %.{ROUND_PRECISION}e' % avg
                else:
                    info += ' %s' % values[k]
            info += ' '  # 最后加个空格，防止中途有别的打印
            self._total_width += len(info)
            if prev_total_width > self._total_width:
                info += (' ' * (prev_total_width - self._total_width))

            if self.target is not None and current >= self.target:
                info += '\n'

            sys.stdout.write(info)
            sys.stdout.flush()

        elif self.verbose == 2:
            if self.target is None or current >= self.target:
                for k in values:
                    info += ' - %s:' % k
                    avg = np.mean(values[k][0] / max(1, values[k][1]))
                    if avg > 1e-3:
                        info += f' %.{ROUND_PRECISION}f' % avg
                    else:
                        info += f' %.{ROUND_PRECISION}e' % avg
                info += '\n'

                sys.stdout.write(info)
                sys.stdout.flush()

        self._last_update = now


class KerasProgbar(Callback):
    ''' keras进度条 '''
    def __init__(self, stateful_metrics=None, interval=None, width=30, **kwargs):
        super(KerasProgbar, self).__init__(**kwargs)
        self.stateful_metrics = _process_stateful_metrics(stateful_metrics)
        self.interval = interval
        self.width = width

    def add_metrics(self, metrics, stateful_metrics=None, add_position=None):
        '''在指定位置插入metrics指标
        '''
        if add_position is None:
            add_position = len(self.params['metrics'])
        metrics = [metrics] if isinstance(metrics, str) else metrics
        if stateful_metrics:
            stateful_metrics = _process_stateful_metrics(stateful_metrics)
            self.stateful_metrics.update(stateful_metrics)
            self.progbar.stateful_metrics.update(stateful_metrics)

        add_metrics = []
        for metric in metrics:
            if metric not in self.params['metrics']:
                add_metrics.append(metric)
        self.params['metrics'] = self.params['metrics'][:add_position] + add_metrics + self.params['metrics'][add_position:]

    def on_train_begin(self, logs=None):
        self.verbose = self.params['verbose']
        self.epochs = self.params['epochs']
        if self.verbose:
            time_start = datetime.now().strftime('%Y-%m-%d %H:%M:%S')
            print('%s - Start Training' % (time_start))

    def on_train_end(self, logs=None):
        if self.verbose:
            time_start = datetime.now().strftime('%Y-%m-%d %H:%M:%S')
            print('\n%s - Finish Training' % (time_start))

    def on_epoch_begin(self, global_step=None, epoch=None, logs=None):
        if self.verbose:
            time_start = datetime.now().strftime('%Y-%m-%d %H:%M:%S')
            print('\n%s - Epoch: %d/%d' % (time_start, epoch+1, self.epochs))
            self.target = self.params['steps']
            self.progbar = Progbar(target=self.target, width=self.width, verbose=self.verbose, stateful_metrics=self.stateful_metrics)
            self.smooth_metric = SmoothMetric(self.interval, self.stateful_metrics)

    def on_batch_end(self, global_step=None, local_step=None, logs=None):
        logs = logs or {}
        log_values = {k:logs[k] for k in self.params['metrics'] if k in logs}
        if self.verbose:
            values = self.smooth_metric.update(local_step+1, log_values)
            self.progbar.update(local_step+1, values)
    
        
class TqdmProgbar(KerasProgbar):
    ''' Tqdm进度条 '''
    def __init__(self, stateful_metrics=None, interval=None, width=None, **kwargs):
        super().__init__(stateful_metrics, interval, width, **kwargs)

    def on_epoch_begin(self, global_step=None, epoch=None, logs=None):
        if self.verbose:
            from tqdm import tqdm
            time_start = datetime.now().strftime('%Y-%m-%d %H:%M:%S')
            print('\n%s - Epoch: %d/%d' % (time_start, epoch+1, self.epochs))
            self.target = self.params['steps']
            self.progbar = tqdm(total=self.params['steps'], desc='Training', dynamic_ncols=False, file=sys.stdout, smoothing=0, ncols=self.width)
        self.smooth_metric = SmoothMetric(self.interval, self.stateful_metrics)

    def on_batch_end(self, global_step=None, local_step=None, logs=None):
        logs_new = self.smooth_values(local_step+1, logs)
        log_values = [(k, logs_new[k]) for k in self.params['metrics'] if k in logs_new]

        # Skip progbar update for the last batch;
        # will be handled by on_epoch_end.
        if self.verbose:
            self.progbar.n = local_step+1
            self.progbar.refresh()
            self.progbar.set_postfix(log_values)

    def on_epoch_end(self, global_step, epoch, logs=None):
        self.progbar.close()
    
    def smooth_values(self, current, values=None):
        '''从Progbar迁移过来'''
        _values = self.smooth_metric.update(current, values)

        logs = collections.OrderedDict()
        for k in _values:
            if isinstance(_values[k], list):
                avg = np.mean(_values[k][0] / max(1, _values[k][1]))
                if abs(avg) > 1e-3:
                    logs[k] = f' %.{ROUND_PRECISION}f' % avg
                else:
                    logs[k] = f' %.{ROUND_PRECISION}e' % avg
            else:
                logs[k] = ' %s' % _values[k]

        return logs


class ProgressBar2Progbar(TqdmProgbar):
    ''' progressbar2进度条 '''
    def on_epoch_begin(self, global_step=None, epoch=None, logs=None):
        if self.verbose:
            import progressbar
            time_start = datetime.now().strftime('%Y-%m-%d %H:%M:%S')
            print('\n%s - Epoch: %d/%d' % (time_start, epoch+1, self.epochs))
            self.target = self.params['steps']
            widgets = [progressbar.SimpleProgress(format='%(value_s)s/%(max_value_s)s'), progressbar.Bar(marker='=', left='[', right=']'), ' ', 
                       progressbar.AdaptiveETA(format='ETA: %(eta)s', format_finished='Time: %(elapsed)s'), ' - ']
            for i, param in enumerate(self.params['metrics']):
                widgets.append(progressbar.Variable(param, precision=7))
                if i < len(self.params['metrics'])-1:
                    widgets.append(' - ')
            self.progbar = progressbar.bar.ProgressBar(min_value=0, max_value=self.params['steps'], widgets=widgets, 
                                                       redirect_stdout=True, redirect_stderr=True)
        self.smooth_metric = SmoothMetric(self.interval, self.stateful_metrics)

    def on_batch_end(self, global_step=None, local_step=None, logs=None):
        logs_new = self.smooth_values(local_step+1, logs or {})
        logs_new = {k:logs_new[k].strip() for k in self.params['metrics'] if k in logs_new}

        # Skip progbar update for the last batch;
        # will be handled by on_epoch_end.
        if self.verbose:
            self.progbar.update(local_step+1, **logs_new)

    def on_epoch_end(self, global_step, epoch, logs=None):
        self.progbar.finish()


class TerminateOnNaN(Callback):
    '''Loss出现NAN停止训练'''
    def on_batch_end(self, global_step, local_step, logs=None):
        logs = logs or {}
        loss = logs.get('loss')
        if loss is not None:
            if np.isnan(loss) or np.isinf(loss):
                log_error('Step %d: Invalid loss, terminating training' % global_step)
                self.trainer.stop_training = True


class History(Callback):
    '''指标历史，默认是fit的返回项, 这里仅记录epoch_end的指标'''
    def on_train_begin(self, logs=None):
        self.epoch = []
        self.history = {}

    def on_epoch_end(self, global_step, epoch, logs=None):
        logs = logs or {}
        self.epoch.append(epoch+1)  # 这里和keras相比+1了
        for k, v in logs.items():
            self.history.setdefault(k, []).append(v)

    def plot(self, ncols=4, subplot_size=(4,3), save_path:str=None, show=True, plot_text=True):
        import matplotlib.pyplot as plt
        import traceback

        nrows = math.ceil(len(self.history) / ncols)
        if nrows <= 1:
            ncols = len(self.history)
        figsize = (subplot_size[0]*ncols, subplot_size[1]*nrows)
        fig, axes = plt.subplots(nrows=nrows, ncols=ncols, sharey=False, squeeze=False, figsize=figsize)
        for i, fea in enumerate(self.history):
            try:
                ax = axes[int(i/ncols), i%ncols]
                ax.plot(self.epoch, self.history[fea], 'bo--', clip_on=False)
                ax.set_title(fea)
                ax.set_xlabel('epoch')
                ax.set_ylabel(fea)
                if plot_text:
                    for x, y in zip(self.epoch, self.history[fea]):
                        ax.text(x, y, f'{set_precision(y)}', ha='center', va='bottom', color='red')
            except:
                log_error(f'Plot `{fea}` error: ' + traceback.format_exc())
        plt.tight_layout()

        if save_path:
            # 保存文件
            save_dir = '/'.join(save_path.split('/')[:-1])
            save_file = save_path.split('/')[-1].split('.')[0] + '.jpg'
            os.makedirs(save_dir, exist_ok=True)
            plt.savefig(os.path.join(save_dir, save_file), dpi=100, bbox_inches='tight')
        if show:
            plt.show()
            plt.close()


class EarlyStopping(Callback):
    '''Stop training策略, 从keras中移植
       使用时候需要保证monitor在logs中，因此如果以valid指标评估需要在EarlyStopping前进行评估并设置logs[monitor]

       :param monitor: str, 监控指标，需要在logs中，默认为'loss', 
       :param min_delta: float, 最小变动，默认为0 
       :param patience: int, 最长等候的次数，默认为0
       :param verbose: int, 是否打印，默认为0表示不打印
       :param mode: str, 控制监控指标monitor的大小方向，默认为'auto', 可选{'auto', 'min', 'max'}
       :param method: str, 控制是按照epoch还是step来计算，默认为'epoch', 可选{'step', 'epoch'}
       :param baseline: None/float, 基线, 默认为None 
       :param restore_best_weights: bool, stopping时候是否恢复最优的权重，默认为False
    '''
    def __init__(self, monitor='loss', min_delta=0, patience=0, verbose=0, mode='auto', method='epoch', baseline=None, restore_best_weights=False, **kwargs):
        super(EarlyStopping, self).__init__(**kwargs)
        assert method in {'step', 'epoch'}, 'Args `method` only support `step` or `epoch`'
        self.method = method  # 默认的epoch和原版一样
        self.monitor = monitor
        self.baseline = baseline
        self.patience = patience  # method=step时候表示最多wait的训练步数
        self.verbose = verbose
        self.min_delta = min_delta
        self.wait = 0
        self.stopped_iteration = 0
        self.restore_best_weights = restore_best_weights
        self.best_weights = None

        if mode not in {'auto', 'min', 'max'}:
            warnings.warn('EarlyStopping mode %s is unknown, fallback to auto mode.' % mode, RuntimeWarning)
            mode = 'auto'

        if mode == 'min':
            self.monitor_op = np.less
        elif mode == 'max':
            self.monitor_op = np.greater
        else:
            self.monitor_op = np.greater if 'acc' in self.monitor else np.less
        self.min_delta = self.min_delta if self.monitor_op == np.greater else -self.min_delta

    def on_train_begin(self, logs=None):
        # Allow instances to be re-used
        self.wait = 0
        self.stopped_iteration = 0
        if self.baseline is not None:
            self.best = self.baseline
        else:
            self.best = np.Inf if self.monitor_op == np.less else -np.Inf

    def on_batch_end(self, global_step, local_step, logs=None):
        if self.method == 'step':
            return self.process(global_step, logs)

    def on_epoch_end(self, global_step, epoch, logs=None):
        if self.method == 'epoch':
            self.process(epoch, logs)

    def process(self, iteration, logs=None):
        logs =  logs or {}
        current = self.get_monitor_value(logs)
        if current is None:
            return

        if self.monitor_op(current - self.min_delta, self.best):
            self.best = current
            self.wait = 0
            if self.restore_best_weights:
                self.best_weights = copy.deepcopy(self.model.state_dict())
        else:
            self.wait += 1
            if self.wait >= self.patience:
                self.stopped_iteration = iteration
                self.trainer.stop_training = True
                # 恢复最优权重
                if self.restore_best_weights:
                    if self.verbose > 0:
                        print('Restoring model weights from the end of the best iteration')
                    self.model.load_state_dict(self.best_weights, strict=True)

    def on_train_end(self, logs=None):
        if self.stopped_iteration > 0 and self.verbose > 0:
            print(f'Iteration {self.stopped_iteration+1}: early stopping\n')

    def get_monitor_value(self, logs):
        monitor_value = logs.get(self.monitor)
        if monitor_value is None:
            warnings.warn('Early stopping conditioned on metric `%s` which is not available. Available metrics are: %s' %
                (self.monitor, ','.join(list(logs.keys()))), RuntimeWarning)
        return monitor_value


class ReduceLROnPlateau(Callback):
    '''当monitor指标不下降时候，降低学习率

    :param monitor: str, 监控指标，需要在logs中，默认为'loss'
    :param factor: float, 权重衰减系数，取值范围(0, 1)，默认为0.1
    :param patience: int, 最长等候的次数，默认为0
    :param method: str, 控制是按照epoch还是step来计算，默认为'epoch', 可选{'step', 'epoch'}
    :param verbose: int, 是否打印，默认为0表示不打印
    :param mode: str, 控制监控指标monitor的大小方向，默认为'auto', 可选{'auto', 'min', 'max'}
    :param min_delta: float, 最小变动，默认为0 
    :param cooldown: float
    :param min_lr: float, 最小学习率
    '''
    def __init__(self, monitor='loss', factor=0.1, patience=10, method='epoch', 
                 verbose=0, mode='auto', min_delta=1e-4, cooldown=0, min_lr=0,
                 **kwargs):
        super(ReduceLROnPlateau, self).__init__(**kwargs)
        assert method in {'step', 'epoch'}, 'Args `method` only support `step` or `epoch`'
        self.method = method  # 默认的epoch和原版一样
        self.monitor = monitor
        if factor >= 1.0:
            raise ValueError('ReduceLROnPlateau does not support a factor >= 1.0.')
        if 'epsilon' in kwargs:
            min_delta = kwargs.pop('epsilon')
            warnings.warn('`epsilon` argument is deprecated and will be removed, use `min_delta` instead.')
        self.factor = factor
        self.min_lr = min_lr
        self.min_delta = min_delta
        self.patience = patience  # method=step时候表示最多wait的训练步数
        self.verbose = verbose
        self.cooldown = cooldown
        self.cooldown_counter = 0  # Cooldown counter.
        self.wait = 0
        self.best = 0
        self.mode = mode
        self.monitor_op = None
        self._reset()

    def _reset(self):
        '''Resets wait counter and cooldown counter.
        '''
        if self.mode not in ['auto', 'min', 'max']:
            warnings.warn('Learning Rate Plateau Reducing mode %s is unknown, fallback to auto mode.' % (self.mode), RuntimeWarning)
            self.mode = 'auto'
        if (self.mode == 'min' or
           (self.mode == 'auto' and 'acc' not in self.monitor)):
            self.monitor_op = lambda a, b: np.less(a, b - self.min_delta)
            self.best = np.Inf
        else:
            self.monitor_op = lambda a, b: np.greater(a, b + self.min_delta)
            self.best = -np.Inf
        self.cooldown_counter = 0
        self.wait = 0

    def on_train_begin(self, logs=None):
        self._reset()

    def on_batch_end(self, global_step, local_step, logs=None):
        if self.method == 'step':
            return self.process(global_step, logs)

    def on_epoch_end(self, global_step, epoch, logs=None):
        if self.method == 'epoch':
            self.process(epoch, logs)

    def process(self, iteration, logs=None):
        logs = logs or {}
        for i, params in enumerate(self.optimizer.param_groups):
            if i == 0:
                logs['lr'] = params["lr"]  # 默认第一个param_group作为lr
            else:
                logs[f'lr_param_group{i}'] = params["lr"]

        current = logs.get(self.monitor)
        if current is None:
            warnings.warn('Reduce LR on plateau conditioned on metric `%s` which is not available. Available metrics are: %s' %
                (self.monitor, ','.join(list(logs.keys()))), RuntimeWarning)

        else:
            if self.in_cooldown():
                self.cooldown_counter -= 1
                self.wait = 0

            if self.monitor_op(current, self.best):
                self.best = current
                self.wait = 0
            elif not self.in_cooldown():
                self.wait += 1
                if self.wait >= self.patience:
                    for params in self.optimizer.param_groups:
                        if 'lr' not in params:
                            continue
                        old_lr = float(params["lr"])
                        if old_lr > self.min_lr:
                            new_lr = old_lr * self.factor
                            new_lr = max(new_lr, self.min_lr)
                            params["lr"] = new_lr
                            if self.verbose > 0:
                                print('\nEpoch %05d: ReduceLROnPlateau reducing learning rate to %s.' % (iteration + 1, new_lr))
                            self.cooldown_counter = self.cooldown
                            self.wait = 0

    def in_cooldown(self):
        return self.cooldown_counter > 0

        
class RemoteMonitor(Callback):
    '''Callback used to stream events to a server.

    :param root: str, url+port
    :param path: str, router
    :param field: str, 字段名
    :param headers: header
    :param send_as_json: bool, 是否以json形式请求，默认为False
    '''
    def __init__(self, root='http://localhost:9000', path='/publish/epoch/end/', field='data',
                 headers=None, send_as_json=False, **kwargs):
        super(RemoteMonitor, self).__init__(**kwargs)
        self.root = root
        self.path = path
        self.field = field
        self.headers = headers
        self.send_as_json = send_as_json
        import requests
        self.requests = requests

    def on_epoch_end(self, global_step, epoch, logs=None):
        logs = logs or {}
        send = {}
        send['epoch'] = epoch
        for k, v in logs.items():
            send[k] = v.item() if isinstance(v, (np.ndarray, np.generic)) else k
        try:
            if self.send_as_json:
                self.requests.post(self.root + self.path, json=send, headers=self.headers)
            else:
                self.requests.post(self.root + self.path, {self.field: json.dumps(send)}, headers=self.headers)
        except self.requests.exceptions.RequestException:
            warnings.warn('Warning: could not reach RemoteMonitor root server at ' + str(self.root))


class Checkpoint(Callback):
    '''保存Checkpoint, 可以每个epoch或者每隔一定的steps保存

    :param save_dir: str, 保存的文件夹，只定义即可按照默认的文件名保存
    :param model_path: str, 模型保存路径(含文件名)，可以使用{epoch}和{step}占位符
    :param optimizer_path: str, 优化器保存路径(含文件名)，可以使用{epoch}和{step}占位符，默认为None表示不保存
    :param scheduler_path: str, scheduler保存路径(含文件名)，可以使用{epoch}和{step}占位符，默认为None表示不保存
    :param steps_params_path: str, 模型训练进度保存路径(含文件名)，可以使用{epoch}和{step}占位符，默认为None表示不保存
    :param method: str, 按照轮次保存还是按照步数保存，默认为'epoch'表示每个epoch保存一次, 可选['epoch', 'step'] 
    :param interval: int, method设置为'step'时候指定每隔多少步数保存模型，默认为100表示每隔100步保存一次
    '''
    def __init__(self, save_dir=None, model_path=None, optimizer_path=None, scheduler_path=None, steps_params_path=None,
                 method='epoch', interval=100, verbose=0, **kwargs):
        super().__init__(**kwargs)
        assert method in {'step', 'epoch'}, 'Args `method` only support `step` or `epoch`'
        self.method = method
        self.save_dir = save_dir  # 保存文件夹
        self.model_path = model_path  # 保存路径，可设置{epoch}{step}{loss}等占位符
        self.optimizer_path = optimizer_path  # 是否保存优化器
        self.scheduler_path = scheduler_path  # 是否保存scheduler
        self.steps_params_path = steps_params_path  # 是否保存训练步数
        self.interval = interval  # method='step'时候生效
        self.verbose = verbose
        self.kwargs = kwargs
    
    def on_epoch_end(self, global_step, epoch, logs=None):
        logs = logs or {}
        if self.method == 'epoch':
            self.process(epoch+1, logs)

    def on_batch_end(self, global_step, local_step, logs=None):
        logs = logs or {}
        if (self.method == 'step') and ((global_step+1) % self.interval == 0):
            self.process(global_step+1, logs)

    def process(self, suffix, logs):
        file_paths = []
        for filepath in [self.save_dir, self.model_path, self.optimizer_path, self.scheduler_path, self.steps_params_path]:
            if filepath:
                filepath = filepath.format(epoch=suffix, **logs) if self.method == 'epoch' else filepath.format(step=suffix, **logs)
            file_paths.append(filepath)

        self.trainer.save_to_checkpoint(*file_paths, verbose=self.verbose, **self.kwargs)


class Evaluator(Checkpoint):
    '''评估并保存最优Checkpoint, 也可以只评估

    :param monitor: str, 监控指标，需要在logs中，默认为'perf'
    :param verbose: int, 是否打印，默认为2表示打印
    :param mode: str, 控制监控指标monitor的大小方向，默认为'auto', 可选{'auto', 'min', 'max'}
    :param method: str, 控制是按照epoch还是step来计算，默认为'epoch', 可选{'step', 'epoch'}
    :param baseline: None/float, 基线, 默认为None 
    :param save_dir: str, 保存的文件夹，只定义即可按照默认的文件名保存
    :param model_path: str, 模型保存路径(含文件名)，可以使用{epoch}和{step}占位符
    :param optimizer_path: str, 优化器保存路径(含文件名)，可以使用{epoch}和{step}占位符，默认为None表示不保存
    :param scheduler_path: str, scheduler保存路径(含文件名)，可以使用{epoch}和{step}占位符，默认为None表示不保存
    :param steps_params_path: str, 模型训练进度保存路径(含文件名)，可以使用{epoch}和{step}占位符，默认为None表示不保存
    :param interval: int, method设置为'step'时候指定每隔多少步数保存模型，默认为100表示每隔100步保存一次
    '''
    def __init__(self, monitor='perf', mode='max', verbose=2, save_dir=None, model_path=None, optimizer_path=None, scheduler_path=None,
                 steps_params_path=None, method='epoch', interval=100, **kwargs):
        super().__init__(save_dir, model_path, optimizer_path, scheduler_path, steps_params_path, method, interval, **kwargs)
        self.monitor = monitor
        assert mode in {'max', 'min'}, 'Compare performance only support `max/min` mode'
        self.mode = mode
        self.verbose = verbose
        self.best_perf = np.inf if mode == 'min' else -np.inf

    def process(self, suffix, logs):
        perf = self.evaluate()
        # 如果evaluate返回的是字典则使用字典，如果返回的是数值则套上{'perf': perf}
        perf = perf if isinstance(perf, dict) else {'perf': perf}
        logs.update(perf)  # 评估的指标后续可能会用到
        
        # 不存在
        if perf.get(self.monitor) is None:
            warnings.warn('Evaluator callback conditioned on metric `%s` which is not available. Available metrics are: %s' %
                (self.monitor, ','.join(list(logs.keys()))), RuntimeWarning)
            return

        # 满足条件
        if ((self.mode == 'max') and (perf[self.monitor] >= self.best_perf)) or ((self.mode == 'min') and (perf[self.monitor] <= self.best_perf)):
            self.best_perf = perf[self.monitor]
            # 保存ckpt
            super().process(suffix, logs)

        if self.verbose != 0:
            print_str = ', '.join([f'{k}: {round(v)}' for k, v in perf.items()])
            print(print_str + f'. best_{self.monitor}: {round(self.best_perf)}')
        
    # 定义评价函数
    def evaluate(self):
        # 需要返回一个字典，且self.monitor在字典key中
        # 如果返回的是一个数值型，则默认使用'perf'作为指标名
        return None


class Logger(Callback):
    '''默认logging
    对于valid/dev和test的日志需要在evaluate之后对log进行赋值，如log['dev_f1']=f1，并在Evaluator之后调用
    若每隔一定steps对验证集评估，则Logger的interval设置成和Evaluater一致或者约数，保证日志能记录到

    :param log_path: str, log文件的保存路径
    :param interval: int, 保存log的间隔
    :param mode: str, log保存的模式, 默认为'a'表示追加
    :param separator: str, 指标间分隔符
    :param level: str, DEBUG/INFO/WARNING/ERROR/CRITICAL，指定log的level
    '''
    def __init__(self, log_path, interval=100, mode='a', separator='\t', level='DEBUG', name='root', **kwargs):
        super(Logger, self).__init__(**kwargs)
        self.log_path = log_path
        self.interval = interval
        self.mode = mode
        self.sep = separator
        self.name = name
        self.level = level

    def on_train_begin(self, logs=None):
        import logging
        level_dict = {'DEBUG': logging.DEBUG, 'INFO': logging.INFO, 'WARNING': logging.WARNING, 
                      'ERROR': logging.ERROR, 'CRITICAL':logging.CRITICAL}
        formatter = logging.Formatter("[%(asctime)s] %(message)s")
        self.logger = logging.getLogger(self.name)
        self.logger.setLevel(level_dict[self.level])
        save_dir = os.path.dirname(self.log_path)
        if save_dir != '':
            os.makedirs(save_dir, exist_ok=True)
        
        fh = logging.FileHandler(self.log_path, self.mode)
        fh.setFormatter(formatter)
        self.logger.addHandler(fh)
        self.logger.info('Start Training'.center(40, '='))

    def on_train_end(self, logs=None):
        self.logger.info('Finish Training'.center(40, '=') + '\n')

    def on_epoch_begin(self, global_step, epoch, logs=None):
        self.logger.info(f'Epoch {epoch+1}'.center(40, '='))

    def on_epoch_end(self, global_step, epoch, logs=None):
        log_str = f'{self.sep}'.join([f'{k}={round(v)}' for k, v in logs.items() if k not in SKIP_METRICS])
        self.logger.info(f'epoch={epoch+1}{self.sep}{log_str}')

    def on_batch_end(self, global_step, local_step, logs=None):
        if (global_step+1) % self.interval == 0:
            log_str = f'{self.sep}'.join([f'{k}={round(v)}' for k, v in logs.items() if k not in SKIP_METRICS])
            self.logger.info(f'step={global_step+1}{self.sep}{log_str}')


class Tensorboard(Callback):
    '''默认Tensorboard
    对于valid/dev和test的Tensorboard需要在evaluate之后对log进行赋值，如log['dev/f1']=f1，并在Evaluator之后调用
    赋值需要分栏目的用'/'进行分隔
    若每隔一定steps对验证集评估，则Tensorboard的interval设置成和Evaluater一致或者约数，保证Tensorboard能记录到

    :param log_dir: str, tensorboard文件的保存路径
    :param interval: int, 保存tensorboard的间隔
    :param prefix: str, tensorboard分栏的前缀，默认为'train'
    '''
    def __init__(self, log_dir, interval=100, prefix='Train', **kwargs):
        super(Tensorboard, self).__init__(**kwargs)
        self.log_dir = log_dir
        self.interval = interval
        self.prefix_step = prefix+'/' if len(prefix.strip()) > 0 else ''  # 控制默认的前缀，用于区分栏目
        self.prefix_epoch = prefix+'_epoch/' if len(prefix.strip()) > 0 else 'Epoch/'  # 控制默认的前缀，用于区分栏目

    def on_train_begin(self, logs=None):
        from tensorboardX import SummaryWriter
        os.makedirs(self.log_dir, exist_ok=True)
        self.writer = SummaryWriter(log_dir=str(self.log_dir))  # prepare summary writer

    def on_epoch_end(self, global_step, epoch, logs=None):
        if hasattr(self.trainer, 'epochs') and self.trainer.epochs > 1:
            self.process(epoch+1, logs, self.prefix_epoch)

    def on_batch_end(self, global_step, local_step, logs=None):
        if (global_step+1) % self.interval == 0:
            self.process(global_step+1, logs, self.prefix_step)

    def process(self, iteration, logs, prefix):
        logs = logs or {}
        for k, v in logs.items():
            if k in SKIP_METRICS:
                continue
            index = k if '/' in k else f"{prefix}{k}"
            self.writer.add_scalar(index, v, iteration)


class WandbCallback(Callback):
    '''从transformers迁移过来
    A :class:`~transformers.TrainerCallback` that sends the logs to `Weight and Biases <https://www.wandb.com/>`__.

    :param interval: int, log的的step间隔
    :param watch (:obj:`str`, `optional` defaults to :obj:`"gradients"`):
        Can be :obj:`"gradients"`, :obj:`"all"` or :obj:`"false"`. Set to :obj:`"false"` to disable gradient
        logging or :obj:`"all"` to log gradients and parameters.
    :param project: str，wandb的project name, 默认为bert4torch
    :param 
    '''
    def __init__(self, project='bert4torch', trial_name=None, run_name=None, watch='gradients', 
                 interval=100, save_code=False, config=None):
        try:
            import wandb
            self._wandb = wandb
        except:
            log_warn("WandbCallback requires wandb to be installed. Run `pip install wandb`.")
            self._wandb = None

        self._initialized = False
        # log outputs
        self.project = project
        if trial_name is None:
            self.trial_name = datetime.now().strftime('%Y-%m-%d %H:%M:%S')
        if run_name is None:
            self.run_name = self.trial_name
        self.watch = watch
        self.interval = interval
        self.save_code = save_code
        self.config = config or {}
        self.run_id = None
        self.metrics = set()

    def define_metric(self, step_metric, logs=None):
        if getattr(self._wandb, "define_metric", None):
            for m in logs.keys():
                if m not in self.metrics:
                    self._wandb.define_metric(name=m, step_metric=step_metric, hidden=True if m in SKIP_METRICS else False)
                    self.metrics.add(m)

    def adjust_logs(self, logs, **kwargs):
        logs_new = {**logs, **kwargs}
        return {k:v for k,v in logs_new.items() if k not in SKIP_METRICS}

    def on_epoch_end(self, global_step, epoch, logs=None):
        if self._wandb is None:
            return
        if hasattr(self.trainer, 'epochs') and self.trainer.epochs > 1:
            self.define_metric('epoch', logs)
            self._wandb.log(self.adjust_logs(logs, epoch=epoch+1))

    def on_batch_end(self, global_step, local_step, logs=None):
        if self._wandb is None:
            return
        
        if (global_step+1) % self.interval == 0:
            self.define_metric('step', logs)
            self._wandb.log(self.adjust_logs(logs, step=global_step+1))
        
    def on_train_begin(self, logs=None):
        if self._wandb is None:
            return

        self._initialized = True
        combined_dict = {**self.trainer.__dict__, **self.config}

        if hasattr(self.model, "config") and self.model.config is not None:
            combined_dict = {**self.model.config, **combined_dict}
        init_args = {}
        if self.trial_name is not None:
            run_name = self.trial_name
            init_args["group"] = self.run_name
        else:
            run_name = self.run_name

        if self._wandb.run is None:
            run = self._wandb.init(project=self.project, name=run_name, save_code=self.save_code, **init_args)
            self.run_id = run.id

        # add config parameters (run may have been created manually)
        self._wandb.config.update(combined_dict, allow_val_change=True)

        # keep track of model topology and gradients, unsupported on TPU
        if self.watch != "false":
            self._wandb.watch(self.model, log=self.watch, log_freq=max(100, self.interval))

    def on_train_end(self, logs=None):
        if self._wandb is None:
            return

        # transformer中的on_log
        self._wandb.finish()
        self._initialized = False


class LambdaCallback(Callback):
    '''lambda表达式
    '''
    def __init__(self, on_epoch_begin=None, on_epoch_end=None, on_batch_begin=None, on_batch_end=None, 
                 on_train_begin=None, on_train_end=None, on_dataloader_end=None, **kwargs):
        super(LambdaCallback, self).__init__(**kwargs)
        self.__dict__.update(kwargs)
        if on_epoch_begin is not None:
            self.on_epoch_begin = on_epoch_begin
        else:
            self.on_epoch_begin = lambda global_step, epoch, logs: None

        if on_epoch_end is not None:
            self.on_epoch_end = on_epoch_end
        else:
            self.on_epoch_end = lambda global_step, epoch, logs: None

        if on_batch_begin is not None:
            self.on_batch_begin = on_batch_begin
        else:
            self.on_batch_begin = lambda global_step, local_step, logs: None

        if on_batch_end is not None:
            self.on_batch_end = on_batch_end
        else:
            self.on_batch_end = lambda global_step, local_step, logs: None

        if on_train_begin is not None:
            self.on_train_begin = on_train_begin
        else:
            self.on_train_begin = lambda logs: None

        if on_train_end is not None:
            self.on_train_end = on_train_end
        else:
            self.on_train_end = lambda logs: None
        
        if on_dataloader_end is not None:
            self.on_dataloader_end = on_train_end
        else:
            self.on_dataloader_end = lambda logs: None


class Summary(Callback):
    '''调用torchinfo的summary
    '''
    def on_train_begin(self, logs=None):
        from torchinfo import summary
        print()
        summary(self.model, input_data=next(iter(self.trainer.train_dataloader))[0])
        print()


class EmailCallback(Callback):
    '''发送Email

    :param mail_receivers: str/list, 收件人邮箱
    :param method: str, 控制是按照epoch还是step来发送邮件，默认为'epoch', 可选{'step', 'epoch'}
    :param interval: int, 发送邮件的的step间隔
    :param mail_host: str, 发件服务器host
    :param mail_user: str, 发件人
    :param mail_pwd: str, smtp的第三方密码
    :param mail_sender: str, 发件人邮箱
    '''
    def __init__(self, mail_receivers, mail_subject='', method='epoch', interval=100, mail_host=None, mail_user=None, mail_pwd=None, mail_sender=None, **kwargs):
        super(EmailCallback, self).__init__(**kwargs)
        self.method = method
        self.interval = interval
        self.mail_receivers = mail_receivers
        self.mail_subject = mail_subject
        self.mail_host = mail_host
        self.mail_user = mail_user
        self.mail_pwd = mail_pwd
        self.mail_sender = mail_sender

    def on_epoch_end(self, global_step, epoch, logs=None):
        if self.method == 'epoch':
            msg = json.dumps({k:f'{round(v)}' for k,v in logs.items() if k not in SKIP_METRICS}, indent=2, ensure_ascii=False)
            subject = f'[INFO] Epoch {epoch+1} performance'
            if self.mail_subject != '':
                subject = self.mail_subject + ' | ' + subject
            self._email(subject, msg)

    def on_batch_end(self, global_step, local_step, logs=None):
        if (self.method == 'step') and ((global_step+1) % self.interval == 0):
            msg = json.dumps({k:f'{round(v)}' for k,v in logs.items() if k not in SKIP_METRICS}, indent=2, ensure_ascii=False)
            subject = f'[INFO] Step {global_step} performance'
            if self.mail_subject != '':
                subject = self.mail_subject + ' | ' + subject
            self._email(subject, msg)

    def on_train_end(self, logs=None):
        msg = json.dumps({k:f'{round(v)}' for k,v in logs.items() if k not in SKIP_METRICS}, indent=2, ensure_ascii=False)
        subject = f'[INFO] Finish training'
        if self.mail_subject != '':
            subject = self.mail_subject + ' | ' + subject
        self._email(subject, msg)

    def _email(self, subject, msg):
        send_email(self.mail_receivers, mail_subject=subject, mail_msg=msg, mail_host=self.mail_host,
                   mail_user=self.mail_user, mail_pwd=self.mail_pwd, mail_sender=self.mail_sender)




from typing import Any, Callable, Dict, List, Optional, Tuple, Union
import time
import os
import traceback
import copy
import functools
from .log import log_info, log_warn, log_error


def spend_time(func):
    '''装饰器，计算函数消耗的时间
    '''
    start = time.time()
    def warpper(*args, **kwargs):
        res = func(*args, **kwargs)
        end = time.time()
        consume = end - start
        start1 = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(start))
        end1 = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(end))
        print(f'{start1} ~ {end1}  spent {consume:.2f}s')
        return res
    return warpper


def send_email(mail_receivers:Union[str,list], mail_subject:str, mail_msg:str="", mail_host:str=None, 
               mail_user:str=None, mail_pwd:str=None, mail_sender:str=None):
    ''' 发送邮件(默认使用笔者自己注册的邮箱，若含敏感信息请使用自己注册的邮箱)

    :param mail_subject: str, 邮件主题
    :param mail_msg: str, 邮件正文
    :param mail_receivers: str/list, 收件人邮箱
    :param mail_host: str, 发件服务器host
    :param mail_user: str, 发件人
    :param mail_pwd: str, smtp的第三方密码
    :param mail_sender: str, 发件人邮箱
    '''
    import smtplib
    from email.mime.text import MIMEText

    mail_host = mail_host or 'smtp.163.com'
    mail_user = mail_user or 'bert4torch'
    mail_pwd = mail_pwd or 'VDSGQEHFXDZOCVEH'
    mail_sender = mail_sender or 'bert4torch@163.com'

    #构造邮件内容
    message = MIMEText(mail_msg,'plain','utf-8')
    message['Subject'] = mail_subject
    message['From'] = mail_sender
    assert isinstance(mail_receivers, (str, tuple, list)), 'Arg `receivers` only receive `str, tuple, list` format'
    message['To'] = mail_receivers if isinstance(mail_receivers, str) else ';'.join(mail_receivers)

    #登录并发送邮件
    try:
        smtpObj = smtplib.SMTP() 
        smtpObj.connect(mail_host, 25)  # 连接到服务器
        smtpObj.login(mail_user, mail_pwd)  # 登录到服务器
        smtpObj.sendmail(mail_sender, mail_receivers, message.as_string())  # 发送
        smtpObj.quit()  # 退出
        log_info('Send email success')
    except smtplib.SMTPException as e:
        log_error('Send email error : '+str(e))
        return str(e)


def monitor_run_by_email(func, mail_receivers:Union[str,list]=None, mail_subject:str=None, mail_host:str=None, 
                         mail_user:str=None, mail_pwd:str=None, mail_sender:str=None):
    """ 通过发邮件的形式监控运行，在程序出现异常的时候发邮件
    """
    @functools.wraps(func)
    def get_except(*args,**kwargs):
        try:
            return func(*args,**kwargs)
        except Exception as e:
            error_msg = traceback.format_exc()
            mail_receivers_ = mail_receivers or kwargs.get('mail_receivers')
            if mail_receivers_ is not None:
                mail_subject_ = mail_subject or kwargs.get('mail_subject') or "[ERROR] " + func.__name__
                mail_host_ = mail_host or kwargs.get('mail_host')
                mail_user_ = mail_user or kwargs.get('mail_user')
                mail_pwd_ = mail_pwd or kwargs.get('mail_pwd')
                mail_sender_ = mail_sender or kwargs.get('mail_sender')
                send_email(mail_receivers_, mail_subject_, error_msg, mail_host=mail_host_, 
                           mail_user=mail_user_, mail_pwd=mail_pwd_, mail_sender=mail_sender_)
            raise e
    return get_except


def email_when_error(receivers:Union[str,list], **configs):
    '''装饰器，异常则发邮件
    Example:
    --------
    @email_when_error(receivers='tongjilibo@163.com')
    def test():
        return 1/0
    test()  # 调用
    '''
    def actual_decorator(func):
        def new_func(*args, **kwargs):
            try:
                res = func(*args, **kwargs)
            except Exception as e:
                error_msg = traceback.format_exc()
                send_email(receivers, func.__name__, error_msg, **configs)
                raise e
            return res
        return new_func
    return actual_decorator


def watch_system_state(log_dir:str, gpu_id_list=None, pid=None):
    '''监控system的状态
    
    :param log_dir: str, tensorboard的地址
    :param gpu_id_list: List[int], 监控的gpu
    :param pid: int, 监控的进程号
    '''
    import psutil
    import pynvml
    from tensorboardX import SummaryWriter

    pynvml.nvmlInit()
    os.makedirs(log_dir, exist_ok=True)
    tb_writer = SummaryWriter(log_dir=str(log_dir))  # prepare summary writer

    pre_time_int = int(time.time())
    pre_time = time.time()
    pid = None or os.getpid()
    p = psutil.Process(pid)
    G = 1024*1024*1024

    pre_read = p.io_counters().read_bytes
    pre_write = p.io_counters().write_bytes
    time_str_init = int(time.time())

    log_info("Watching System Info")
    while True:
        time.sleep(1)
        now_time = time.time()
        time_str = int(now_time) - time_str_init
        p = psutil.Process(pid)

        # CPU使用情况
        tb_writer.add_scalar(f"CPU_pid_{pid}/cpu_percent", p.cpu_percent(interval=0.5), time_str)

        # 内存使用情况
        tb_writer.add_scalar(f"Memory_pid_{pid}/memory_percent", p.memory_percent(), time_str)
        tb_writer.add_scalar(f"Memory_pid_{pid}/RSS G_byte", p.memory_info().rss/G, time_str)
        tb_writer.add_scalar(f"Memory_pid_{pid}/VMS G_byte", p.memory_info().vms/G, time_str)

        # 进程IO信息
        data = p.io_counters()
        tb_writer.add_scalar(f"IO_pid_{pid}/read M_byte", data.read_bytes/1024, time_str)
        tb_writer.add_scalar(f"IO_pid_{pid}/write M_byte", data.write_bytes/1024, time_str)
        if (time_str - pre_time_int)%5 == 0:
            tb_writer.add_scalar(f"IO_pid_{pid}/readRate M_byte_s", (data.read_bytes - pre_read)/(now_time-pre_time)/1024, time_str)
            tb_writer.add_scalar(f"IO_pid_{pid}/writeRate M_byte_s", (data.write_bytes - pre_write)/(now_time-pre_time)/1024, time_str)
            pre_read = data.read_bytes
            pre_write = data.write_bytes
            pre_time = now_time

        tesorboard_info_list = []
        if gpu_id_list is None:
            deviceCount = pynvml.nvmlDeviceGetCount()
            device_list = [i for i in range(deviceCount)]
        else:
            device_list = gpu_id_list

        tesorboard_info_list = []
        for i in device_list:
            tesorboard_info = {}
            handle = pynvml.nvmlDeviceGetHandleByIndex(i)
            tesorboard_info['gpu_name'] = pynvml.nvmlDeviceGetName(handle)
            meminfo = pynvml.nvmlDeviceGetMemoryInfo(handle)
            tesorboard_info['gpu_mem_used'] = meminfo.used/G
            UTIL = pynvml.nvmlDeviceGetUtilizationRates(handle)
            tesorboard_info['gpu_gpu_util'] = UTIL.gpu
            tesorboard_info['gpu_mem_util'] = UTIL.memory
            tesorboard_info_list.append(copy.deepcopy(tesorboard_info))
        for tesorboard_info,i in zip(tesorboard_info_list, device_list):
            tb_writer.add_scalar(f'GPU{i}/gpu_mem_used unit_G', tesorboard_info['gpu_mem_used'], time_str)
            tb_writer.add_scalar(f'GPU{i}/gpu_gpu_util', tesorboard_info['gpu_gpu_util'], time_str)
            tb_writer.add_scalar(f'GPU{i}/gpu_mem_util', tesorboard_info['gpu_mem_util'], time_str)

    tb_writer.close()
    pynvml.nvmlShutdown()

    return


import numpy as np
import torch
from torch import nn, Tensor
from typing import Any, Callable, Dict, List, Optional, Tuple, Union
import os
import random
from .log import log_info, log_warn, log_error


def seed_everything(seed:int=None):
    '''固定seed
    
    :param seed: int, 随机种子
    '''
    max_seed_value = np.iinfo(np.uint32).max
    min_seed_value = np.iinfo(np.uint32).min

    if (seed is None) or not (min_seed_value <= seed <= max_seed_value):
        seed = random.randint(np.iinfo(np.uint32).min, np.iinfo(np.uint32).max)
    log_info(f"Global seed set to {seed}")
    os.environ["PYTHONHASHSEED"] = str(seed)
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    torch.backends.cudnn.benchmark = False
    torch.backends.cudnn.deterministic = True
    return seed


def print_trainable_parameters(module:nn.Module):
    '''打印可训练的参数量'''
    trainable_params = 0
    all_param = 0
    for _, param in module.named_parameters():
        num_params = param.numel()
        # if using DS Zero 3 and the weights are initialized empty
        if num_params == 0 and hasattr(param, "ds_numel"):
            num_params = param.ds_numel

        all_param += num_params
        if param.requires_grad:
            trainable_params += num_params
    log_info(f"trainable params: {trainable_params} || all params: {all_param} || trainable%: {100 * trainable_params / all_param}")


def get_parameter_device(parameter):
    '''获取device, 从transformers包迁移过来'''
    try:
        return next(parameter.parameters()).device
    except StopIteration:
        # For nn.DataParallel compatibility in PyTorch 1.5

        def find_tensor_attributes(module: nn.Module) -> List[Tuple[str, Tensor]]:
            tuples = [(k, v) for k, v in module.__dict__.items() if torch.is_tensor(v)]
            return tuples

        gen = parameter._named_members(get_members_fn=find_tensor_attributes)
        first_tuple = next(gen)
        return first_tuple[1].device


class DottableDict(dict):
    '''支持点操作符的字典'''
    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)
        self.__dict__ = self
    def allowDotting(self, state=True):
        if state:
            self.__dict__ = self
        else:
            self.__dict__ = dict()
KwargsConfig = DottableDict


class JsonConfig:
    '''读取配置文件并返回可.操作符的字典'''
    def __new__(self, json_path, encoding='utf-8'):
        import json
        return DottableDict(json.load(open(json_path, "r", encoding=encoding)))


def auto_set_cuda_devices(best_num: Optional[int] = None) -> str:
    '''
    这段代码是一个名为 auto_set_cuda_devices 的函数，它接受一个可选的整数参数 best_num。该函数用于自动设置环境变量 CUDA_VISIBLE_DEVICES，以便在多个 GPU 设备中选择最佳的 GPU 设备。
    首先，该函数检查环境变量 CUDA_VISIBLE_DEVICES 是否已经设置。如果已经设置，则发出警告并返回当前设置的值。
    如果 best_num 等于 -1，则将环境变量 CUDA_VISIBLE_DEVICES 设置为 -1 并返回。
    接下来，该函数尝试使用 nvidia-smi 命令查询 GPU 设备的信息。如果命令不存在，则发出警告并将环境变量 CUDA_VISIBLE_DEVICES 设置为 -1 并返回。
    如果未指定 best_num，则从环境变量 LOCAL_WORLD_SIZE 中获取值。然后将其转换为整数。
    接下来，该函数解析 nvidia-smi 命令的输出，并计算每个 GPU 设备的得分。得分由 GPU 利用率和可用内存计算得出。
    最后，该函数选择得分最高的 best_num 个 GPU 设备，并将其索引作为字符串连接起来。然后将环境变量 CUDA_VISIBLE_DEVICES 设置为该字符串并返回。
    '''
    import subprocess

    # 无需重复设置
    if "CUDA_VISIBLE_DEVICES" in os.environ:
        log_warn("Environment variable `CUDA_VISIBLE_DEVICES` has already been set")
        return os.environ["CUDA_VISIBLE_DEVICES"]

    if best_num == -1:
        log_info(f"SET CUDA_VISIBLE_DEVICES=-1")
        os.environ["CUDA_VISIBLE_DEVICES"] = "-1"
        return "-1"

    try:
        p = subprocess.Popen(
            ["nvidia-smi",
             "--query-gpu=index,utilization.gpu,memory.total,memory.free",
             "--format=csv,noheader,nounits"],
            stdout=subprocess.PIPE)
    except FileNotFoundError:
        log_error("`nvidia-smi` not exists")
        os.environ["CUDA_VISIBLE_DEVICES"] = "-1"
        return "-1"

    if best_num is None:
        best_num = os.environ.get("LOCAL_WORLD_SIZE", 1)
    best_num = int(best_num)

    stdout, _ = p.communicate()
    outputs = stdout.decode("utf-8")
    lines = outputs.split(os.linesep)

    scores = []
    for item in lines:
        item_arr = item.split(",")
        if len(item_arr) != 4:
            continue
        gpu_score = 100 - int(item_arr[1].strip())
        memory_score = 100 * (int(item_arr[3].strip()) / int(item_arr[2].strip()))
        score = gpu_score + memory_score
        scores.append(score)

    best_num = min(best_num, len(scores))
    topk_idx = (-np.asarray(scores)).argsort()[:best_num]

    topk_idx_str = ",".join(map(str, topk_idx))
    log_info(f"SET CUDA_VISIBLE_DEVICES={topk_idx_str}")
    os.environ["CUDA_VISIBLE_DEVICES"] = topk_idx_str

    return topk_idx_str




#! -*- coding: utf-8 -*-
# continue pretrain

from bert4torch.snippets import sequence_padding
import torch.nn as nn
import torch
import torch.optim as optim
from torch.utils.data import DataLoader
import torch
from bert4torch.snippets import IterDataset, DottableDict
from bert4torch.tokenizers import Tokenizer
from glob import glob


# 基本参数
args = DottableDict()
args.data_path = 'E:/Github/MedicalGPT/data/pretrain/**/*.txt'
args.batch_size = 10
args.max_seq_length = 128

dict_path = 'E:/pretrain_ckpt/bert/chinese_L-12_H-768_A-12/vocab.txt'
tokenizer = Tokenizer(dict_path, do_lower_case=True)

# 加载数据集, 数据量较大使用IterDataset
class MyDataset(IterDataset):
    @staticmethod
    def load_data(filenames):
        """加载数据，并尽量分为不超过maxlen的句子
        """
        D = []
        for filename in filenames:
            with open(filename, encoding='utf-8') as f:
                for l in f:
                    input_ids = tokenizer.encode(l)[0]
                    if len(input_ids) > 0:
                        input_ids = input_ids[:-1]
                    if len(D) + len(input_ids) > args.max_seq_length-1:  # +当前输入超长的话，则返回之前的累计输入
                        yield D
                        D = input_ids
                    else:
                        D.extend(input_ids)

def collate_fn(batch_token_ids):
    batch_token_ids = torch.tensor(sequence_padding(batch_token_ids, value=tokenizer.pad_token_id), dtype=torch.long)
    return [batch_token_ids], batch_token_ids

file_list = glob(args.data_path, recursive=True) * 100
train_dataloader = DataLoader(MyDataset(file_list), batch_size=args.batch_size, collate_fn=collate_fn) 

count = 0
for batch in train_dataloader:
    count += 1
    if count % 100 == 0:
        print(count)
print('done')


from torch import nn
import torch
from torch4keras.snippets import DottableDict, metric_mapping, get_parameter_device, log_info, log_warn, log_error
from torch4keras.snippets import print_trainable_parameters, colorful, monitor_run_by_email, load_checkpoint, save_checkpoint
from torch4keras.callbacks import KerasProgbar, SmoothMetricsCallback, TqdmProgbar, ProgressBar2Progbar, Callback, CallbackList, History
from collections import OrderedDict
from typing import Union
from inspect import isfunction
import os
import json
import math


class Trainer:
    '''Trainer, 传入Module实例

    :param module: None/nn.Module，nn.Module()的模型实例
    '''
    def __init__(self, module:nn.Module=None):
        super(Trainer, self).__init__()
        self.initialize(module)
    
    def initialize(self, module:nn.Module=None):
        # 传入Module实例方式
        if module is not None:
            assert isinstance(module, nn.Module), 'Args `module` only support nn.Module format'
            self.module = module

        self.global_step, self.local_step, self.total_steps, self.batch_step = 0, 0, 0, 0
        self.epoch, self.steps_per_epoch, self.train_dataloader = 0, None, None
        self.resume_step, self.resume_epoch = 0, 0
        self.retain_graph = False  # loss.backward()是否保留计算图
        self.move_to_model_device = True  # 自动把tensor转到model所在的device
        self.log_first_step = False  # 是否打印第一个step的数据
        self.criterion = None  # criterion
        self.optimizer = None  # optimizer
        self.scheduler = None  # scheduler
        self.callbacks = []  # 所有的Callbacks，如果fit中不传入，则默认为[progbarlogger, smoothmetrics, history]三项
        self.run_callbacks = True  # 是否运行Callbacks，目前主要是在DDP模式下运用
        self.loss2metrics = True  # 把loss_detail打印在进度条的metrics上
        # add_module(self)  # 增加nn.Module的成员方法

    def compile(self, loss=None, optimizer=None, scheduler=None, clip_grad_norm=None, mixed_precision=False, metrics=None, 
                grad_accumulation_steps=1, progbar_config=None, smooth_metrics_config=None, **kwargs):
        '''complile: 定义loss, optimizer, metrics等参数
        
        :param loss: loss
        :param optimizer: 优化器
        :param scheduler: lr_scheduler
        :param clip_grad_norm: bool, 是否使用梯度裁剪, 默认为False
        :param mixed_precision: bool, 是否使用混合精度，默认为False
        :param metrics: str/List[str]/dict, 训练过程中需要打印的指标, loss相关指标默认会打印, 目前支持accuracy, 也支持自定义metric，形式为{key: func}
        :param grad_accumulation_steps: int, 梯度累积步数，默认为1
        :param bar: str, 使用进度条的种类，从kwargs中解析，默认为keras, 可选keras, tqdm, progressbar2
        :param progbar_config: 进度条的配置，默认是对整个epoch计算均值指标
            bar: str, 默认为keras
            stateful_metrics: List[str], 表示不使用指标平滑仅进行状态记录的metric，指标抖动会更加明显，默认为None表示使用指标平滑
            width: int, keras进度条下表示进度条的长度
        :param smooth_metrics_config: 指标平滑的配置，默认为None表示采取默认平滑设置；传入False表示不使用平滑
            stateful_metrics: List[str], 表示不使用指标平滑仅进行状态记录的metric，指标抖动会更加明显，默认为None表示使用指标平滑
            interval: int, 表示指标平滑时候的累计步数，默认为100

        :return: None
        '''
        self.criterion = loss or self.criterion  # loss
        self.optimizer = optimizer or self.optimizer  # 优化器
        self.scheduler = scheduler or self.scheduler  # lr_scheduler
        self.clip_grad_norm = clip_grad_norm  # 梯度裁剪
        self.grad_accumulation_steps = grad_accumulation_steps  # 梯度累积

        # 混合精度
        assert mixed_precision in {True, False, 'fp16', 'bf16'}
        self.mixed_precision = 'fp16' if mixed_precision is True else mixed_precision
        if self.mixed_precision:
            self.autocast = torch.cuda.amp.autocast
            self.scaler = torch.cuda.amp.GradScaler()

        # 训练过程观测的指标
        self.metrics = OrderedDict({'loss': None})
        if metrics is None:
            metrics = []
        elif isinstance(metrics, (str, dict)) or isfunction(metrics):
            metrics = [metrics]
        for metric in metrics:
            # 字符类型，目前仅支持accuracy
            if isinstance(metric, str) and metric != 'loss':
                self.metrics[metric] = None
            # 字典形式 {metric: func}
            elif isinstance(metric, dict):
                self.metrics.update(metric)
            # 函数形式，key和value都赋值metric
            elif isfunction(metric):
                self.metrics.update({metric: metric})
            else:
                raise ValueError('Args metrics only support `String, Dict, Callback, List[String, Dict, Callback]` format')

        # 进度条参数
        self.progbar_config = progbar_config or {'bar': 'keras', 'stateful_metrics': None}
        self.progbar_config.update({k:v for k, v in kwargs.items() if k in ['bar', 'stateful_metrics', 'width']})  # 直接传参
        self.progbar_config['bar'] = self.progbar_config.get('bar', 'keras')
        assert self.progbar_config['bar'] in {'keras', 'tqdm', 'progressbar2'}, \
            f'Args `bar`={self.progbar_config["bar"]} illegal, only support `keras, tqdm, progressbar2`'

        # smooth_metrics参数: 默认平滑
        if smooth_metrics_config is False:  # compile时传入False, 表示不使用平滑
            self.smooth_metrics_config = None
        else:
            self.smooth_metrics_config = smooth_metrics_config or {}
            self.smooth_metrics_config.update({k:v for k, v in kwargs.items() if k in ['stateful_metrics', 'interval', 'verbose']})  # 直接传参

        # 其他参数设置
        for key, value in kwargs.items():
            if hasattr(self, key):
                setattr(self, key, value)

    def print_trainable_parameters(self):
        '''打印可训练的参数量'''
        print_trainable_parameters(self.unwrap_model())

    @property
    def device(self) -> torch.device:
        '''获取model所在的device'''
        if hasattr(self, '_device'):
            return self._device
        return get_parameter_device(self.unwrap_model())

    @device.setter
    def device(self, value):
        '''允许修改self.device'''
        self._device = value

    def _move_to_model_device(self, inputs):
        '''遍历并转移到model.device上（递归）'''
        if self.move_to_model_device:
            if isinstance(inputs, torch.Tensor) and hasattr(self, 'device') and (inputs.device != self.device):
                inputs = inputs.to(self.device)
            elif isinstance(inputs, (tuple, list)):
                inputs = list(inputs) if isinstance(inputs, tuple) else inputs
                for i, ts in enumerate(inputs):
                    inputs[i] = self._move_to_model_device(ts)
            elif isinstance(inputs, dict):
                for k, v in inputs.items():
                    inputs[k] = self._move_to_model_device(v)
        return inputs

    def _log_first_step(self, resume_step, train_X, train_y):
        '''打印第一个step的数据'''
        if self.log_first_step and self.verbose and (self.global_step == resume_step):
            print(colorful('[Train_data]: ', color='green'), + train_X)
            print(colorful('[Label]: ', color='green'), + train_y)

    def _forward(self, *inputs, **input_kwargs):
        '''调用模型的forward，方便下游继承的时候可以自定义使用哪个模型的forward
        '''
        return self._argparse_forward(self.unwrap_model(), *inputs, **input_kwargs)

    def _argparse_forward(self, model, *inputs, **input_kwargs):
        '''调用模型的forward
        如果传入了网络结构module，则调用module的forward；如果是继承方式，则调用自身的forward
        这里声明为staticmethod的话，使用add_trainer会有问题
        '''
        if (len(inputs)==1) and isinstance(inputs[0], (tuple,list)):  # 防止([])嵌套
            inputs = inputs[0]
        
        if isinstance(inputs, torch.Tensor):  # tensor不展开
            return model.forward(inputs, **input_kwargs)
        elif isinstance(inputs, (tuple, list)):
            return model.forward(*inputs, **input_kwargs)
        else:
            return model.forward(inputs, **input_kwargs)

    def train_step(self, train_X, train_y):
        ''' Perform a training step on a batch of inputs. '''
        # 计算loss
        if self.mixed_precision:
            with self.autocast(dtype=torch.float16 if self.mixed_precision=='fp16' else torch.bfloat16):
                output = self._forward(train_X)
                loss_detail = self.criterion(output, train_y)
        else:
            output = self._forward(train_X)
            loss_detail = self.criterion(output, train_y)

        # 整理loss
        if isinstance(loss_detail, torch.Tensor):
            loss = loss_detail
            loss_detail = {}
        elif isinstance(loss_detail, dict):
            loss = loss_detail['loss']  # 还存在其他loss，仅用于打印
            del loss_detail['loss']
        elif isinstance(loss_detail, (tuple, list)):
            loss = loss_detail[0]
            loss_detail = {f'loss{i}':v for i, v in enumerate(loss_detail[1:], start=1)}
        else:
            raise ValueError('Return loss only support `Tensor/dict/tuple/list` format')

        # 梯度累积
        loss = loss / self.grad_accumulation_steps if self.grad_accumulation_steps > 1 else loss

        # loss backward
        loss = self.loss_backward(loss)
        loss_detail = {k: (v.item() if isinstance(v, torch.Tensor) else v) / self.grad_accumulation_steps for k, v in loss_detail.items()}
        return output, loss, loss_detail

    def loss_backward(self, loss):
        '''loss.backward'''
        self.scale_before_step = 0
        if self.mixed_precision:  # 混合精度
            self.scale_before_step = self.scaler.get_scale()
            self.scaler.scale(loss).backward(retain_graph=self.retain_graph)
        else:
            loss.backward(retain_graph=self.retain_graph)
        return loss
    
    def step(self):
        '''参数更新'''
        skip_scheduler = False
        # 混合精度
        if self.mixed_precision:
            self.scaler.unscale_(self.optimizer)
            if self.clip_grad_norm is not None:  # 梯度裁剪
                torch.nn.utils.clip_grad_norm_(self.unwrap_model().parameters(), self.clip_grad_norm)
            self.scaler.step(self.optimizer)
            self.scaler.update()
            skip_scheduler = self.scaler.get_scale() != self.scale_before_step
        else:
            if self.clip_grad_norm is not None:  # 梯度裁剪
                torch.nn.utils.clip_grad_norm_(self.unwrap_model().parameters(), self.clip_grad_norm)
            self.optimizer.step()

        self.optimizer.zero_grad()  # 清梯度
        if (self.scheduler is not None) and not skip_scheduler:
            if isinstance(self.scheduler, (tuple, list)):
                for scheduler in self.scheduler:
                    scheduler.step()
            else:
                self.scheduler.step()

    def _prepare_inputs(self, train_dataloader, steps_per_epoch, epochs, verbose):
        '''对fit的输入进行类型检查并置为成员变量'''
        if not hasattr(train_dataloader, '__len__'):
            assert steps_per_epoch is not None, 'Either train_dataloader has attr `__len__` or steps_per_epoch is not None'
        if steps_per_epoch is None:
            self.steps_per_epoch = math.ceil(len(train_dataloader) // self.grad_accumulation_steps)
        else:
            self.steps_per_epoch = steps_per_epoch
        self.batch_size = train_dataloader.batch_size
        self.epochs = epochs
        self.total_steps = self.steps_per_epoch * epochs
        self.train_dataloader = train_dataloader  # 设置为成员变量，可由外部的callbacks进行修改
        self.train_dataloader_iter = iter(self.train_dataloader)  # 循环epoch时不重生成
        self.verbose = self.verbose if hasattr(self, 'verbose') else verbose

    def _prepare_callbacks(self, callbacks):
        '''callbacks设置'''
        if callbacks is None:
            callbacks = []
        elif isinstance(callbacks, Callback):
            callbacks = [callbacks]
        for callback in callbacks:
            assert isinstance(callback, Callback), "Args `callbacks` only support Callback() inputs"

        history = History()
        callbacks_ = []

        # 进度条
        progbarlogger = None
        if self.verbose:
            if self.progbar_config['bar'] == 'keras':
                progbarlogger = KerasProgbar(**self.progbar_config)
            elif self.progbar_config['bar'] == 'tqdm':
                progbarlogger = TqdmProgbar(**self.progbar_config)
            elif self.progbar_config['bar'] == 'progressbar2':
                progbarlogger = ProgressBar2Progbar(**self.progbar_config)
            else:
                progbarlogger = KerasProgbar(**self.progbar_config)
            callbacks_.append(progbarlogger)

        # 指标平滑
        if self.smooth_metrics_config is not None:
            if any([isinstance(i, SmoothMetricsCallback) for i in callbacks]):
                # 用户自定的callbacks中包含了SmoothMetricsCallback
                log_warn(f'SmoothMetricsCallback already in use and args `smooth_metrics_config` will be ignored')
            else:
                callbacks_.append(SmoothMetricsCallback(**self.smooth_metrics_config))

            # 检查指标平滑的设置和后续callback的设置的interval是不是一致
            smooth_callback = [callback for callback in callbacks_+callbacks if isinstance(callback, SmoothMetricsCallback)][0]
            for callback in callbacks_+callbacks:
                if hasattr(callback, 'interval') and (callback != smooth_callback) and (callback.interval is not None) and \
                    (callback.interval != smooth_callback.interval):
                    log_warn(f'{type(callback).__name__}.interval={callback.interval} while SmoothMetricsCallback.interval={smooth_callback.interval}')

        callbacks_  += callbacks + [history]
        self.callbacks = CallbackList(callbacks_, run_callbacks=self.run_callbacks)
        callback_trainer = self
        callback_model = self.unwrap_model()
        params = {
            'epochs': self.epochs,
            'steps': self.steps_per_epoch,
            'verbose': self.verbose,
            'metrics': [i for i in self.metrics.keys() if isinstance(i, str)],
        }
        self.callbacks.set_all(trainer=callback_trainer, model=callback_model, optimizer=self.optimizer, scheduler=self.scheduler, params=params)
        callback_trainer.stop_training = False  # 在EarlyStopping中会重新设置
        return history, callback_trainer, progbarlogger

    def _prepare_nextbatch(self):
        '''准备下一个batch数据'''
        # 循环dataloader, 不要试用itertools的cycle，遇到过变量不释放的问题
        try:
            batch = next(self.train_dataloader_iter)
            self.batch_step += 1
        except StopIteration:
            self.callbacks.on_dataloader_end()  # 适用于数据量较大时，动态读取文件并重新生成self.train_dataloader的情况，如预训练
            # DDP训练时候为了避免每个epoch样本一致，修改随机种子
            if isinstance(self.train_dataloader.sampler, torch.utils.data.distributed.DistributedSampler) and \
                hasattr(self.train_dataloader.sampler, 'set_epoch'):
                self.train_dataloader.sampler.set_epoch(self.epoch)
            self.train_dataloader_iter = iter(self.train_dataloader)  # shuffle=True时候，其实顺序也重新生成了
            self.batch_step = 0
            batch = next(self.train_dataloader_iter)

        batch = self._move_to_model_device(batch)
        return batch

    @monitor_run_by_email
    def fit(self, train_dataloader, steps_per_epoch=None, epochs=1, callbacks=None, verbose=1, **kwargs):
        '''模型训练
        
        :param train_dataloader: Dataloader, 训练数据集
        :param steps_per_epoch: int, 每个epoch训练的steps，默认为None表示自行计算 
        :param epochs: int, 训练的轮次, 默认为1
        :param callbacks: Callback/List[Callback], 回调函数，可调用预制的Callback或者自定义，默认为None 
        :param verbose: int, 是否打印，默认为1表示打印
        :return: History
        '''
        # 输入处理
        self._prepare_inputs(train_dataloader, steps_per_epoch, epochs, verbose)

        # 准备callbacks
        history, callback_trainer, progbarlogger  = self._prepare_callbacks(callbacks)

        #       epoch: 当前epoch
        # global_step: 当前全局训练步数
        #  local_step: 当前epoch内的训练步数，不同epoch中相同local_step对应的batch数据不一定相同，在steps_per_epoch=None时相同
        #  batch_step: 在dataloader中的index，不同epoch中相同的bti对应的batch数据一般相同，除非重新生成dataloader
        self.callbacks.on_train_begin()
        for epoch in range(self.resume_epoch, epochs):
            self.epoch = epoch
            # resume_step：判断local_step的起点，以及进度条的起始位置
            resume_step = self.resume_step if epoch==self.resume_epoch else 0
            self.callbacks.on_epoch_begin(self.global_step, self.epoch)
            if self.verbose:
                progbarlogger.seen = resume_step  # 这里设置进度条的seen，在callbacks中也会修改
            
            for local_step in range(resume_step, self.steps_per_epoch):
                self.local_step = local_step
                self.global_step = self.epoch * self.steps_per_epoch + self.local_step
                logs = self._log_init()
                self.callbacks.on_batch_begin(self.global_step, self.local_step, logs)

                # forward和backward
                self.unwrap_model().train()  # 设置为train模式
                tr_loss, tr_loss_detail = 0, {}
                for _ in range(self.grad_accumulation_steps):
                    train_X, train_y = self._prepare_nextbatch()  # 获取下一个batch的训练数据
                    self._log_first_step(resume_step, train_X, train_y)  # log第一个step
                    output, loss, loss_detail = self.train_step(train_X, train_y)
                    self.callbacks.on_train_step_end()
                    tr_loss += loss.item()
                    for k, v in loss_detail.items():
                        tr_loss_detail[k] = tr_loss_detail.get(k, 0) + v
                # TODO: 理论上梯度累积时需对output和train_y进行合并，主要是为了metric_mapping计算的准确
                
                # 参数更新
                self.step()

                # 添加loss至log打印
                logs.update(dict({'loss': tr_loss}, **tr_loss_detail))
                if self.verbose and self.loss2metrics and (self.global_step == resume_step):
                    # 把loss_detail添加到进度条metrics中
                    progbarlogger.add_metrics(list(tr_loss_detail.keys()), add_position=1)
                    
                # 添加metrics至log打印
                for metric, func in self.metrics.items():
                    perf = metric_mapping(metric, func, output, train_y)  # 内置的一些accuracy指标
                    if perf is not None:
                        if isfunction(metric):  # 直接传入回调函数(无key)
                            if self.verbose and (self.global_step == resume_step):
                                progbarlogger.add_metrics(list(perf.keys()))
                            logs.update(perf)
                        elif isinstance(metric, str):  # 直接传入回调函数(有key)
                            logs[metric] = perf

                self.callbacks.on_batch_end(self.global_step, self.local_step, logs)

            self.callbacks.on_epoch_end(self.global_step, self.epoch, logs)
            # TerminateOnNaN、EarlyStopping等停止训练策略
            if callback_trainer.stop_training:
                break
        self.callbacks.on_train_end(logs)
        return history

    def _log_init(self):
        '''获取batch_size，主要是用于callback中的BaseLogger和Callback
        '''
        logs = {}

        # 添加lr
        try:
            logs['lr'] = self.optimizer.param_groups[0]["lr"]
        except:
            pass
        return logs

    @torch.no_grad()
    def predict(self, *inputs, **input_kwargs):
        '''模型预测，调用forward()'''
        self.unwrap_model().eval()
        inputs = self._move_to_model_device(inputs)
        input_kwargs = self._move_to_model_device(input_kwargs)
        return self._forward(*inputs, **input_kwargs)
        
    def load_steps_params(self, save_path):
        '''导入训练过程参数
        
        :param save_path: str, 训练过程参数保存路径
        '''
        step_params = torch.load(save_path)
        self.resume_step = step_params['resume_step'] 
        self.resume_epoch = step_params['resume_epoch']
        return step_params

    def save_steps_params(self, save_path):
        '''保存训练过程参数

        :param save_path: str, 训练过程参数保存路径
        '''
        step_params = {'resume_step': (self.local_step+1) % self.steps_per_epoch, 
                       'resume_epoch': self.epoch + (self.local_step+1) // self.steps_per_epoch}
        save_dir = os.path.dirname(save_path)
        os.makedirs(save_dir, exist_ok=True)
        torch.save(step_params, save_path)

    def load_weights(self, load_path:Union[str,tuple,list], strict:bool=True, mapping:dict=None):
        '''加载模型权重, 支持加载权重文件list

        :param save_path: str/tuple/list, 权重加载路径
        :param strict: bool, torch.load()是否严格加载
        :param mapping: dict, 指定key的映射
            1. mapping=None, 表示按照模型自身结构加载，一般加载finetune后使用save_weights()保存出来的权重
            2. mapping自定义，根据用户自定义mapping来加载权重
        '''
        if isinstance(load_path, (tuple, list)):
            strict = False  # 加载多个权重文件时候，strict设置为False
        elif isinstance(load_path, str):
            load_path = [load_path]
        else:
            raise ValueError('Args `load_path` only support str/tuple/list format')
        
        mapping = mapping or dict()
        for load_path_i in load_path:
            state_dict = load_checkpoint(load_path_i)
            for k in list(state_dict.keys()):
                if k in mapping:
                    state_dict[mapping[k]] = state_dict.pop(k)
            self.unwrap_model().load_state_dict(state_dict, strict=strict)

    def save_weights(self, save_path:str, mapping:dict=None, trainable_only:bool=False):
        '''保存模型权重

        :param save_path: str, 权重保存路径
        :param mapping: dict, 指定key的映射
            1. mapping=None, 表示按照模型自身结构的key来保存，后续可直接load_weights()加载
            2. mapping自定义，根据用户自定义mapping来保存权重
        :param trainable_only: bool, 指定仅保存可训练参数
        '''
        state_dict = self.unwrap_model().state_dict()
        trainable_parameters = set(p for p,v in self.unwrap_model().named_parameters() if v.requires_grad)
        
        mapping = mapping or dict()
        for k in list(state_dict.keys()):
            # 只保存可训练的模型部分
            if trainable_only and (k not in trainable_parameters):
                continue
            if k in mapping:
                state_dict[mapping[k]] = state_dict.pop(k)
        
        save_checkpoint(state_dict, save_path)
        if trainable_only:
            params_all = sum(p.numel() for p in self.unwrap_model().parameters())
            params_trainable = sum(p.numel() for p in self.unwrap_model().parameters() if p.requires_grad)
            ratio = params_trainable/params_all*100
            log_info(f"Only trainable parameters saved and occupy {params_trainable}/{params_all}={ratio:.2f}%")

    def save_pretrained(self, save_path:str, weight_map:dict=None, mapping:dict=None):
        '''按照预训练模型的key来保存模型, 可供transformers包加载'''
        state_dict = dict()
        for name, child in self.unwrap_model().named_children():
            if (name != '') and hasattr(child, 'save_pretrained'):
                tmp = child.save_pretrained(save_path, weight_map, mapping, write_to_disk=False)
                state_dict.update(tmp if tmp else {})
            else:
                state_dict.update({f'{name}.{k}': v for k,v in child.state_dict().items()})
        save_checkpoint(state_dict, save_path)
    
    def resume_from_checkpoint(self, save_dir:str=None, model_path:str=None, optimizer_path:str=None, scheduler_path:str=None, 
                               steps_params_path:str=None, mapping:dict=None, verbose:int=0, strict:bool=True, **kwargs):
        '''同时加载模型、优化器、训练过程参数

        :param save_dir: str, 保存目录
        :param model_path: str, 模型文件路径
        :param optimizer_path: str, 优化器文件路径
        :param scheduler_path: str, scheduler文件路径
        :param steps_params_path: str, 训练过程参数保存路径
        :param mapping: dict, 模型文件的mapping
        '''
        model_path = model_path or os.path.join(save_dir, 'model.pt')
        optimizer_path = optimizer_path or os.path.join(save_dir, 'optimizer.pt')
        scheduler_path = scheduler_path or os.path.join(save_dir, 'scheduler.pt')
        steps_params_path = steps_params_path or os.path.join(save_dir, 'steps_params.pt')

        verbose_str = ''
        # 加载模型权重
        if model_path:
            self.load_weights(model_path, strict=strict, mapping=mapping)
            verbose_str += f'Model weights successfuly resumed from {model_path}\n'
        # 加载优化器，断点续训使用
        if optimizer_path:
            state_dict = torch.load(optimizer_path, map_location='cpu')
            self.optimizer.load_state_dict(state_dict)
            verbose_str += f'Optimizer successfuly resumed from {optimizer_path}\n'
        # 加载优化器，断点续训使用
        if scheduler_path:
            state_dict = torch.load(scheduler_path, map_location='cpu')
            self.scheduler.load_state_dict(state_dict)
            verbose_str += f'Scheduler successfuly resumed from {scheduler_path}\n'
        # 加载训练进度参数，断点续训使用
        if steps_params_path:
            self.load_steps_params(steps_params_path)
            verbose_str += f'Steps_params successfuly resumed from {steps_params_path}'
        if verbose == 1:
            print(verbose_str)

    def save_to_checkpoint(self, save_dir:str=None, model_path:str=None, optimizer_path:str=None, scheduler_path:str=None, 
                           steps_params_path:str=None, mapping:dict=None, trainable_only:bool=False, verbose:int=0, **kwargs):
        '''同时保存模型、优化器、训练过程参数、scheduler

        :param save_dir: str, 保存目录
        :param model_path: str, 模型文件路径
        :param optimizer_path: str, 优化器文件路径
        :param scheduler_path: str, scheduler文件路径
        :param steps_params_path: str, 训练过程参数保存路径
        :param mapping: dict, 模型文件的mapping
        :param trainable_only
        '''
        model_path = model_path or os.path.join(save_dir, 'model.pt')
        optimizer_path = optimizer_path or os.path.join(save_dir, 'optimizer.pt')
        scheduler_path = scheduler_path or os.path.join(save_dir, 'scheduler.pt')
        steps_params_path = steps_params_path or os.path.join(save_dir, 'steps_params.pt')

        verbose_str = ''
        if model_path:
            self.save_weights(model_path, mapping=mapping, trainable_only=trainable_only)
            verbose_str += f'Model weights successfuly saved to {model_path}\n'
        if optimizer_path:
            save_dir = os.path.dirname(optimizer_path)
            os.makedirs(save_dir, exist_ok=True)
            torch.save(self.optimizer.state_dict(), optimizer_path)
            verbose_str += f'Optimizer successfuly saved to {optimizer_path}\n'
        if scheduler_path and (self.scheduler is not None):
            save_dir = os.path.dirname(scheduler_path)
            os.makedirs(save_dir, exist_ok=True)
            torch.save(self.scheduler.state_dict(), scheduler_path)
            verbose_str += f'Scheduler successfuly saved to {scheduler_path}\n'
        if steps_params_path:
            self.save_steps_params(steps_params_path)
            verbose_str += f'Steps_params successfuly saved to {steps_params_path}'
        if verbose == 1:
            print(verbose_str)

    def unwrap_model(self):
        '''返回nn.Module模块
        '''
        if isinstance(self, nn.Module): return self
        return self.module if hasattr(self, 'module') else self


Trainer.compile_training_components = Trainer.compile


class TrainerDP(nn.DataParallel, Trainer):
    '''DataParallel模式使用多gpu的方法, 
    1) 父类顺序颠倒也会出问题
    2) 使用方式和nn.DataParallel一致，TrainerDP(net, *args, **kwargs)来使用
    '''
    def __init__(self, *args, **kwargs):
        Trainer.__init__(self)
        nn.DataParallel.__init__(self, *args, **kwargs)


class TrainerDDP(nn.parallel.DistributedDataParallel, Trainer):
    '''DistributedDataParallel模式使用多gpu的方法,
    1) 父类顺序颠倒也会出问题
    2) 使用方式和DistributedDataParallel一致，TrainerDDP(net, *args, **kwargs)来使用
    '''
    def __init__(self, *args, master_rank=0, **kwargs):
        Trainer.__init__(self)
        nn.parallel.DistributedDataParallel.__init__(self, *args, **kwargs)

        # 默认仅对master_rank=0打印信息
        assert isinstance(master_rank, (int, list, tuple)), 'Args `master_rank` only supoorts int, list, tuple'
        if isinstance(master_rank, int):
            master_rank = [master_rank]
        self.verbose = (torch.distributed.get_rank() in master_rank)
    

def add_trainer(obj, include=None, exclude=None, verbose=0, replace_func=False):
    '''为nn.Module添加Triner对应的方法'''
    if isinstance(obj, (Trainer, TrainerDP, TrainerDDP)):
        log_warn('obj is not a Trainer object')
        return obj
    elif not isinstance(obj, nn.Module):
        log_warn('obj is not a nn.Module object')
        return obj

    if include is None:
        include = set()
    elif isinstance(include, str):
        include = set([include])
    elif isinstance(include, (tuple, list)):
        include = set(include)
    else:
        raise TypeError(f'Arg `include` only receive str/list format, not {type(include)}')

    if exclude is None:
        exclude = set()
    elif isinstance(exclude, (tuple, list)):
        exclude = set(exclude)

    import types
    added_funcs = []
    for k in dir(Trainer):
        set_func = False
        if k in include:  # 必须包含的
            set_func = True
        elif k in exclude:  # 必须屏蔽的
            pass
        elif k.startswith('__') and k.endswith('__'):  # 内部函数不执行
            pass
        elif hasattr(obj, k):  # 如果下游重新定义，则不继
            if replace_func:
                set_func = True
        else:
            set_func = True

        if set_func and eval(f'isfunction(Trainer.{k})'):
            exec(f'obj.{k} = types.MethodType(Trainer.{k}, obj)')
            added_funcs.append(k)
    obj.initialize()  # 这里初始化会得到一些其他的成员变量，不可缺省

    if verbose and (len(added_funcs) > 0):
        log_info(f'Already add `{",".join(added_funcs)}` method')
    return obj


def add_module(obj, include=None, exclude=None, verbose=0, replace_func=False):
    '''为Trainer增加nn.Module的方法
    方便外部访问, 如obj.parameters()可以直接访问到obj.module.parameters()
    '''
    if isinstance(obj, nn.Module):
        return obj
    elif not isinstance(obj, Trainer):
        log_warn('obj is not a Trainer object')
        return obj
    elif not isinstance(obj.unwrap_model(), nn.Module):
        log_warn('obj.unwrap_model() is not a nn.Module object')
        return obj
    
    if include is None:
        include = set()
    elif isinstance(include, str):
        include = set([include])
    elif isinstance(include, (tuple, list)):
        include = set(include)
    else:
        raise TypeError(f'Arg `include` only receive str/list format, not {type(include)}')

    if exclude is None:
        exclude = set()
    elif isinstance(exclude, (tuple, list)):
        exclude = set(exclude)


    import types
    added_funcs = []
    for k in dir(obj.unwrap_model()):
        set_func = False
        if k in include:  # 必须包含的
            set_func = True
        elif k in exclude:  # 必须屏蔽的
            pass
        elif k.startswith('__') and k.endswith('__'):
            pass
        elif hasattr(obj, k):  # 如果下游重新定义，则不继
            if replace_func:
                set_func = True
        else:
            set_func = True
        
        if set_func and eval(f'isinstance(obj.unwrap_model().{k}, types.MethodType)'):
            exec(f'obj.{k} = obj.unwrap_model().{k}')
            added_funcs.append(k)

    if verbose and (len(added_funcs) > 0):
        log_info(f'Already add `{",".join(added_funcs)}` method')
    return obj


class AccelerateTrainer(Trainer):
    '''accelerate来训练'''
    def __init__(self, module: nn.Module, **configs):
        super().__init__(module)
        from accelerate import Accelerator
        accelerator = Accelerator(**configs)
        self.model = accelerator.prepare(module)
        self.accelerator = accelerator
        self.device = accelerator.device
        self.verbose = 1 if accelerator.is_local_main_process else 0
        log_warn('AcclerateTrainer may not be compatible with several callbacks, you may use custom callbacks instead.')
    
    def compile(self, *args, **kwargs):
        super().compile(*args, **kwargs)
        self.optimizer, self.scheduler, self.criterion = self.accelerator.prepare(self.optimizer, self.scheduler, self.criterion)

    def _prepare_inputs(self, *args):
        super()._prepare_inputs(*args)
        self.train_dataloader = self.accelerator.prepare(self.train_dataloader)
        self.train_dataloader_iter = iter(self.train_dataloader)

    def prepare(self, *args, **kwargs):
        '''调用acclerate的prepare，如在外面评估时候需要对dev_dataloader使用'''
        return self.accelerator.prepare(*args, **kwargs)

    def unwrap_model(self):
        '''返回nn.Module模块'''
        unwrap_model = self.accelerator.unwrap_model(self.model)
        if isinstance(unwrap_model, nn.Module): return unwrap_model
        return unwrap_model.module if hasattr(unwrap_model, 'module') else unwrap_model

    def loss_backward(self, loss):
        self.accelerator.backward(loss)
        return loss


class DeepSpeedTrainer(Trainer):
    '''deepspeed来训练'''
    def __init__(self, module, config_path):
        super().__init__(module)
        self.model = module
        self.config = DottableDict(json.load(open(config_path)))
        self.config['steps_per_print'] = self.config.get('steps_per_print', 1e9)  # 默认不打印，防止进度条打印问题

    def compile(self, *args, log_level='warning', inference=False, master_rank=0, **kwargs):
        super().compile(*args, **kwargs)
        import deepspeed
        from deepspeed.utils import logger as ds_logger
        import logging
        log_levels = {
            "debug": logging.DEBUG,
            "info": logging.INFO,
            "warning": logging.WARNING,
            "error": logging.ERROR,
            "critical": logging.CRITICAL,
        }

        ds_logger.setLevel(log_levels.get(log_level, logging.WARNING))

        if inference:
            # only Z3 makes sense for the inference
            log_warn("ZeRO inference only makes sense with ZeRO Stage 3")
            self.optimizer, self.scheduler = None, None
            model_parameters = None
        else:
            model_parameters = list(filter(lambda p: p.requires_grad, self.model.parameters()))
        
        kwargs = {
            "model": self.model,  # deepspeed的forward默认是计算到loss输出的
            "model_parameters": model_parameters,
            "config_params": self.config,
            "optimizer": self.optimizer,
            "lr_scheduler": self.scheduler,
        }
        if self.config.get('zero_optimization', {}).get('offload_optimizer', {}).get('device') == 'cpu':
            kwargs.pop('optimizer')
            if self.optimizer is not None:
                self.optimizer = None
                log_warn('You may not use custom optimizer when offload_optimizer=`cpu`')
        self.deepspeed_engine, self.optimizer, _, self.scheduler = deepspeed.initialize(**kwargs)
        self.verbose = 1 if self.deepspeed_engine.local_rank == master_rank else 0

    def unwrap_model(self):
        # 执行deepspeed_engine的forward
        return self.deepspeed_engine

    def loss_backward(self, loss):
        self.deepspeed_engine.backward(loss)
        return loss
    
    def step(self):
        self.deepspeed_engine.step()

    def resume_from_checkpoint(self, *args, **kwargs):
        return self.deepspeed_engine.load_checkpoint(*args, **kwargs)

    def save_to_checkpoint(self, *args, **kwargs):
        return self.deepspeed_engine.save_checkpoint(*args, **kwargs)



from torch import nn

# torch2.0以后nn.Module自带compile方法
# 和model.compile()冲突，这里把nn.Module.compile重命名为compile_torch
if hasattr(nn.Module, 'compile'):
    nn.Module.compile_torch = nn.Module.compile
    del nn.Module.compile


from typing import Any, Callable, Dict, List, Optional, Tuple, Union
import os
import functools
import logging
import datetime


def colorful(obj, color="yellow", display_type="plain"):
    '''
    # 彩色输出格式：
    # 设置颜色开始 ：\033[显示方式;前景色;背景色m
    # 说明：
    # 前景色            背景色           颜色
    # ---------------------------------------
    # 30                40              黑色
    # 31                41              红色
    # 32                42              绿色
    # 33                43              黃色
    # 34                44              蓝色
    # 35                45              紫红色
    # 36                46              青蓝色
    # 37                47              白色
    # 显示方式           意义
    # -------------------------
    # 0                终端默认设置
    # 1                高亮显示
    # 4                使用下划线
    # 5                闪烁
    # 7                反白显示
    # 8                不可见
    '''
    color_dict = {"black":"30", "red":"31", "green":"32", "yellow":"33",
                    "blue":"34", "purple":"35","cyan":"36",  "white":"37"}
    display_type_dict = {"plain":"0","highlight":"1","underline":"4",
                "shine":"5","inverse":"7","invisible":"8"}
    s = str(obj)
    color_code = color_dict.get(color,"")
    display  = display_type_dict.get(display_type,"")
    out = '\033[{};{}m'.format(display,color_code)+s+'\033[0m'
    return out 


def log_level(string:str, level:Union[int, str]=0, verbose:int=1):
    '''在字符串前面加上有颜色的[INFO][WARNING][ERROR]字样'''
    if level in {0, 'i', 'info', 'INFO'}:
        res = log_info(string, verbose)
    elif level in {1, 'w', 'warn', 'warning', 'WARN', 'WARNING'}:
        res = log_warn(string, verbose)
    elif level == {2, 'e', 'error', 'ERROR'}:
        res = log_error(string, verbose)
    elif level == 1:
        res = string
    return res
info_level_prefix = log_level


def log_info(string:str, verbose:int=1):
    '''[INFO]: message, 绿色前缀'''
    res = colorful('[INFO]', color='green') + ' ' + string.strip()
    if verbose != 0:
        print(res)
    return res


def log_warn(string:str, verbose:int=1):
    '''[WARNING]: message, 黄色前缀'''
    res = colorful('[WARNING]', color='yellow') + ' ' + string.strip()
    if verbose != 0:
        print(res)
    return res


def log_error(string:str, verbose:int=1):
    '''[ERROR]: message, 红色前缀'''
    res = colorful('[ERROR]', color='red') + ' ' + string.strip()
    if verbose != 0:
        print(res)
    return res


@functools.lru_cache(None)
def log_warn_once(string:str, verbose=1):
    ''' 单次warning '''
    return log_warn(string, verbose)


@functools.lru_cache(None)
def print_once(string:str):
    '''单次打印'''
    print(string)


class SimpleStreamFileLogger(object):
    '''同时在和命令行和日切片中log（简单场景下可用）
    1. 会按照天进行切片，适用于部署一些服务的时候长时间log日志的情况
    2. 缺点是log时候不能定位到文件和行号

    :param log_path: str, log文件的地址, 如'/home/logs/demo.log'
    :param date_format: str, date在路径中的位置，可选subdir|prefix|suffix, subdir会以'/home/logs/20231218/demo.log'保存
    '''
    def __init__(self, log_path:str, date_format:str='subdir', level:str='DEBUG', 
                 format:str="[%(asctime)s][%(levelname)s] %(message)s"):
        self.log_path = log_path
        self.date_format = date_format
        self.level = level
        self.format = format

        self.save_dir = os.path.dirname(log_path)
        if self.save_dir:
            os.makedirs(self.save_dir, exist_ok=True)
        self.file_name = os.path.basename(log_path)
        self.initialize()

    def initialize(self):
        self.date = str(datetime.date.today()).replace('-','')
        # 保存路径
        if self.date_format == 'subdir':
            self.file = os.path.join(self.save_dir, self.date, self.file_name)
        elif self.date_format == 'prefix':
            self.file = os.path.join(self.save_dir, self.date + '_' + self.file_name)
        elif self.date_format == 'suffix':
            self.file = os.path.join(self.save_dir, self.file_name + '_' + self.date)
        else:
            raise ValueError('Args `date_format` only support subdir|prefix|suffix.')

        if (save_dir := os.path.dirname(self.file)) != '':
            os.makedirs(save_dir, exist_ok=True)

        # 创建日志器
        self.log = logging.getLogger(self.file_name)

        # 日志器默认级别
        level_dict = {'DEBUG': logging.DEBUG, 'INFO': logging.INFO, 'WARNING': logging.WARNING, 
                      'ERROR': logging.ERROR, 'CRITICAL':logging.CRITICAL}
        self.log.setLevel(level_dict[self.level])

        # 日志格式器
        self.formatter = logging.Formatter(self.format)
        self.log.handlers.clear()

        # 终端流输出
        stream_handle = logging.StreamHandler()
        self.log.addHandler(stream_handle)
        stream_handle.setLevel(level_dict[self.level])
        stream_handle.setFormatter(self.formatter)

        # 文件流输出
        file_handle = logging.FileHandler(filename=self.file, mode='a', encoding='utf-8')
        self.log.addHandler(file_handle)
        file_handle.setLevel(level_dict[self.level])
        file_handle.setFormatter(self.formatter)
        return self.log
    
    def reinitialize(self):
        if str(datetime.date.today()).replace('-','') != self.date:
            del self.log
            self.initialize()

    def info(self, text):
        self.reinitialize()
        self.log.info(text)
    
    def warn(self, text):
        self.reinitialize()
        self.log.warning(text)
    
    def warning(self, text):
        self.reinitialize()
        self.log.warning(text)

    def error(self, text):
        self.reinitialize()
        self.log.error(text)


class LoggerHandler(logging.Logger):
    '''同时在文件中和命令行中log（推荐）
    1. 适用于部署一些服务的时候长时间log日志的情况，可按照文件大小，日期进行切片
    2. 可以定位到对应的文件和行号

    :param log_path: str, log到文件的路径
    :param handles: str/tuple/list/set, handles的类型，默认为('StreamHandler', 'FileHander')
    - StreamHandler 命令行输出
    - FileHander 文件数据
    - RotatingFileHandler 按照文件size切片的输出
    - TimedRotatingFileHandler 按照时间的切片输出

    :handle_config: dict, handles使用到的config
    :param level: str, DEBUG/INFO/WARNING/ERROR/CRITICAL，指定log的level
    '''
    def __init__(self, log_path:str=None, handles=None, handle_config=None, name:str='root', 
                 level:str='DEBUG', format:str="[%(asctime)s][%(filename)s][line:%(lineno)d][%(levelname)s] %(message)s"):
        super().__init__(name)
        # 日志器默认级别
        level_dict = {'DEBUG': logging.DEBUG, 'INFO': logging.INFO, 'WARNING': logging.WARNING, 
                      'ERROR': logging.ERROR, 'CRITICAL':logging.CRITICAL}
        self.setLevel(level_dict[level])

        if handles is None:
            handles = ('StreamHandler', 'FileHander')
        elif isinstance(handles, str):
            handles = [handles]
        assert isinstance(handles, (tuple, set, list)), 'Args `handles` should be str/tuple/list/set'

        # 日志格式器
        formatter = logging.Formatter(format)
        self.handlers.clear()

        # 终端流输出
        if 'StreamHandler' in handles:
            stream_handle = logging.StreamHandler()
            self.addHandler(stream_handle)
            stream_handle.setLevel(level_dict[level])
            stream_handle.setFormatter(formatter)

        # 文件流输出
        if log_path is None:
            return
        if (dirname := os.path.dirname(log_path)) != '':
            os.makedirs(dirname, exist_ok=True)
        if 'FileHander' in handles:
            file_handle = logging.FileHandler(filename=log_path, mode='a', encoding='utf-8')
        elif 'RotatingFileHandler' in handles:
            if handle_config is None:
                handle_config = {'maxBytes': 1024*1024*10, 'backupCount': 30, 'encoding': 'utf-8'}
            from logging.handlers import RotatingFileHandler
            file_handle = RotatingFileHandler(log_path, **handle_config)
        elif 'TimedRotatingFileHandler' in handles:
            if handle_config is None:
                handle_config = {'when': 'midnight', 'backupCount': 30, 'encoding': 'utf-8'}
            from logging.handlers import TimedRotatingFileHandler
            file_handle = TimedRotatingFileHandler(log_path, **handle_config)
        else:
            return
        self.addHandler(file_handle)
        file_handle.setLevel(level_dict[level])
        file_handle.setFormatter(formatter)



from torch import nn
from torch4keras.trainer import *


class BaseModel(Trainer, nn.Module):
    '''BaseModel, 使用继承的方式来使用
    '''
    def __init__(self, *args, **kwargs):
        nn.Module.__init__(self)
        Trainer.__init__(self, *args, **kwargs)
        
BaseModelDP = TrainerDP  # 使用方式和nn.DataParallel一致
BaseModelDDP = TrainerDDP  # 使用方式和DistributedDataParallel一致


#! -*- coding: utf-8 -*-

from setuptools import setup, find_packages

with open("README.md", "r", encoding='utf-8') as fh:
    long_description = fh.read()

setup(
    name='torch4keras',
    version='v0.1.8',
    description='Use torch like keras',
    long_description=long_description,
    long_description_content_type="text/markdown",
    license='Apache License',
    url='https://github.com/Tongjilibo/torch4keras',
    author='Tongjilibo',
    install_requires=[],
    packages=find_packages()
)